"""API Documentation

Comprehensive AI Toolkit for Multi Modal Learning and Cross-Platform Robotics.

This Module imports the following contents from the sub-module.
"""

from .adapter import Adapter
from .bot import Bot
from .config import ConfigModel
from .dependencies import Depends
from .event import Event, MessageEvent
from .plugin import Plugin
from . import libcore

__all__ = [
    "Adapter",
    "Bot",
    "ConfigModel",
    "Depends",
    "Event",
    "MessageEvent",
    "Plugin",
    "libcore",
]




from iamai import Plugin


class Halloiamai(Plugin):
    async def handle(self) -> None:
        await self.event.reply("Hello, iamai!")

    async def rule(self) -> bool:
        if self.event.adapter.name != "mirai":
            return False
        if self.event.type != "FriendMessage":
            return False
        return self.event.message.get_plain_text().lower() == "hello"



from iamai import Plugin
from iamai.exceptions import GetEventTimeout


class Weather(Plugin):
    async def handle(self) -> None:
        args = self.event.get_plain_text().split(" ")

        if len(args) >= 2:
            await self.event.reply(await self.get_weather(args[1]))
            return

        try:
            city_event = await self.event.ask("请输入想要查询天气的城市：", timeout=10)
        except GetEventTimeout:
            return
        else:
            await self.event.reply(await self.get_weather(city_event.get_plain_text()))

    async def rule(self) -> bool:
        if self.event.adapter.name != "onebot11":
            return False
        if self.event.type != "message":
            return False
        return self.event.message.startswith("天气") or self.event.message.startswith(
            "weather"
        )

    @staticmethod
    async def get_weather(city):
        if city not in ["北京", "上海"]:
            return "你想查询的城市暂不支持！"
        return f"{city}的天气是..."



from iamai import Plugin
from iamai.adapter.apscheduler import scheduler_decorator


@scheduler_decorator(
    trigger="interval", trigger_args={"seconds": 5}, override_rule=True
)
class Send(Plugin):
    async def handle(self) -> None:
        if self.state is None:
            self.state = 0
        await self.bot.get_adapter("bililive").send(str(self.state))
        self.state += 1

    async def rule(self) -> bool:
        return False



"""iamai 插件。

所有 iamai 插件的基类。所有用户编写的插件必须继承自 `Plugin` 类。
"""

import inspect
from abc import ABC, abstractmethod
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    NoReturn,
    Optional,
    Tuple,
    Type,
    cast,
    final,
)
from typing_extensions import Annotated, get_args, get_origin

from .config import ConfigModel
from .dependencies import Depends
from .event import Event
from .exceptions import SkipException, StopException
from .typing import ConfigT, EventT, StateT
from .utils import is_config_class

if TYPE_CHECKING:
    from .bot import Bot

__all__ = ["Plugin", "PluginLoadType"]


class PluginLoadType(Enum):
    """Plugins loaded types."""

    DIR = "dir"
    NAME = "name"
    FILE = "file"
    CLASS = "class"


class Plugin(ABC, Generic[EventT, StateT, ConfigT]):
    """Base class for all iamai plugins.

    Attributes:
        event: The event currently being processed by this plugin.
        priority: The priority of the plugin. The smaller the number, the higher the priority. The default is 0.
        block: Whether to prevent the propagation of events after the plug-in is executed. ``True`` means blocking.
        __plugin_load_type__: Plugin load type, automatically set by iamai, reflects how this plugin is loaded.
        __plugin_file_path__: ``None`` when the plugin loading type is ``PluginLoadType.CLASS``,
            Otherwise, the location of the Python module in which the plugin is defined.
    """

    priority: ClassVar[int] = 0
    block: ClassVar[bool] = False

    # Cannot use ClassVar because PEP 526 does not allow it
    Config: Type[ConfigT]

    __plugin_load_type__: ClassVar[PluginLoadType]
    __plugin_file_path__: ClassVar[Optional[str]]

    if TYPE_CHECKING:
        event: EventT
    else:
        event = Depends(Event)

    def __init_state__(self) -> Optional[StateT]:
        """Initialize plugin state."""

    def __init_subclass__(
        cls,
        config: Optional[Type[ConfigT]] = None,
        init_state: Optional[StateT] = None,
        **_kwargs: Any,
    ) -> None:
        """Initialize subclasses.

        Args:
            config: Configuration class.
            init_state: initial state.
        """
        super().__init_subclass__()

        orig_bases: Tuple[type, ...] = getattr(cls, "__orig_bases__", ())
        for orig_base in orig_bases:
            origin_class = get_origin(orig_base)
            if inspect.isclass(origin_class) and issubclass(origin_class, Plugin):
                try:
                    _event_t, state_t, config_t = cast(
                        Tuple[EventT, StateT, ConfigT], get_args(orig_base)
                    )
                except ValueError:  # pragma: no cover
                    continue
                if (
                    config is None
                    and inspect.isclass(config_t)
                    and issubclass(config_t, ConfigModel)
                ):
                    config = config_t  # pyright: ignore
                if (
                    init_state is None
                    and get_origin(state_t) is Annotated
                    and hasattr(state_t, "__metadata__")
                ):
                    init_state = state_t.__metadata__[0]  # pyright: ignore

        if not hasattr(cls, "Config") and config is not None:
            cls.Config = config
        if cls.__init_state__ is Plugin.__init_state__ and init_state is not None:
            cls.__init_state__ = lambda _: init_state  # type: ignore

    @final
    @property
    def name(self) -> str:
        """plugin class name."""
        return self.__class__.__name__

    @final
    @property
    def bot(self) -> "Bot":
        """bot object."""
        return self.event.adapter.bot  # pylint: disable=no-member

    @final
    @property
    def config(self) -> ConfigT:
        """plugin configuration."""
        default: Any = None
        config_class = getattr(self, "Config", None)
        if is_config_class(config_class):
            return getattr(
                self.bot.config.plugin,
                config_class.__config_name__,
                default,
            )
        return default

    @final
    def stop(self) -> NoReturn:
        """Stop propagation of current events."""
        raise StopException

    @final
    def skip(self) -> NoReturn:
        """Skips itself and continues propagation of the current event."""
        raise SkipException

    @property
    def state(self) -> StateT:
        """plugin status."""
        return self.bot.plugin_state[self.name]

    @state.setter
    @final
    def state(self, value: StateT) -> None:
        self.bot.plugin_state[self.name] = value

    @abstractmethod
    async def handle(self) -> None:
        """Method to handle events. iamai will call this method when the ``rule()`` method returns ``True``. Each plugin must implement this method."""
        raise NotImplementedError

    @abstractmethod
    async def rule(self) -> bool:
        """Method to match the event. When the event is processed, this method will be called in sequence according to the priority of the plugin. When this method returns ``True``, the event will be handed over to this plugin for processing. Each plugin must implement this method.

        .. note::
            It is not recommended to implement event processing directly in this method. Please leave the specific processing of events to the ``handle()`` method.
        """
        raise NotImplementedError



"""DingTalk 协议适配器。

本适配器适配了钉钉企业自建机器人协议。
协议详情请参考：[钉钉开放平台](https://open.dingtalk.com/document/robots/robot-overview)。
"""

import base64
import hashlib
import hmac
import time
from typing import Any, Dict, Literal, Union

import aiohttp
from aiohttp import web

from iamai.adapter import Adapter
from iamai.log import logger

from .config import Config
from .event import DingTalkEvent
from .exceptions import NetworkError
from .message import DingTalkMessage

__all__ = ["DingTalkAdapter"]


class DingTalkAdapter(Adapter[DingTalkEvent, Config]):
    """钉钉协议适配器。"""

    name: str = "dingtalk"
    Config = Config

    app: web.Application
    runner: web.AppRunner
    site: web.TCPSite

    session: aiohttp.ClientSession

    async def startup(self) -> None:
        """创建 aiohttp Application。"""
        self.app = web.Application()
        self.app.add_routes([web.post(self.config.url, self.handler)])

        self.session = aiohttp.ClientSession()

    async def run(self) -> None:
        """运行 aiohttp 服务器。"""
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        self.site = web.TCPSite(self.runner, self.config.host, self.config.port)
        await self.site.start()

    async def shutdown(self) -> None:
        """清理 aiohttp AppRunner。"""
        await self.session.close()
        await self.site.stop()
        await self.runner.cleanup()

    async def handler(self, request: web.Request) -> web.Response:
        """处理 aiohttp 服务器的接收。

        Args:
            request: aiohttp 服务器的 `Request` 对象。
        """
        if "timestamp" not in request.headers or "sign" not in request.headers:
            logger.error("Illegal http header, incomplete http header")
        elif (
            abs(int(request.headers["timestamp"]) - time.time() * 1000) > 60 * 60 * 1000
        ):
            logger.error(
                f'Illegal http header, timestamp: {request.headers["timestamp"]}'
            )
        elif request.headers["sign"] != self.get_sign(request.headers["timestamp"]):
            logger.error(f'Illegal http header, sign: {request.headers["sign"]}')
        else:
            try:
                dingtalk_event = DingTalkEvent(adapter=self, **(await request.json()))
            except Exception as e:
                self.bot.error_or_exception("Request parsing error:", e)
                return web.Response()
            await self.handle_event(dingtalk_event)
        return web.Response()

    def get_sign(self, timestamp: str) -> str:
        """计算签名。

        Args:
            timestamp: 时间戳。

        Returns:
            签名。
        """
        hmac_code = hmac.new(
            self.config.app_secret.encode("utf-8"),
            f"{timestamp}\n{self.config.app_secret}".encode(),
            digestmod=hashlib.sha256,
        ).digest()
        return base64.b64encode(hmac_code).decode("utf-8")

    async def send(
        self,
        webhook: str,
        conversation_type: Literal["1", "2"],
        msg: Union[str, Dict[str, Any], DingTalkMessage],
        at: Union[None, Dict[str, Any], DingTalkMessage] = None,
    ) -> Dict[str, Any]:
        """发送消息。

        Args:
            webhook: Webhook 网址。
            conversation_type: 聊天类型，"1" 表示单聊，"2" 表示群聊。
            msg: 消息。
            at: At 对象，仅在群聊时生效，默认为空。

        Returns:
            钉钉服务器的响应。

        Raises:
            TypeError: 传入参数类型错误。
            ValueError: 传入参数值错误。
            NetworkError: 调用 Webhook 地址时网络错误。
        """
        if isinstance(msg, DingTalkMessage):
            pass
        elif isinstance(msg, dict):
            msg = DingTalkMessage.raw(msg)
        elif isinstance(msg, str):
            msg = DingTalkMessage.text(msg)
        else:
            raise TypeError(
                f"msg must be str, Dict or DingTalkMessage, not {type(msg)!r}"
            )

        if at is not None:
            if isinstance(at, DingTalkMessage):
                if at.type == "at":
                    pass
                else:
                    raise ValueError(f'at.type must be "at", not {at.type}')
            elif isinstance(at, dict):
                at = DingTalkMessage.raw(at)
            else:
                raise TypeError(f"at must be Dict or DingTalkMessage, not {type(at)!r}")

        data: Union[Dict[str, Any], DingTalkMessage]
        if conversation_type == "1":
            data = msg
        elif conversation_type == "2":
            if at is None:
                data = {"msgtype": msg.type, **msg.model_dump()}
            else:
                data = {"msgtype": msg.type, **msg.model_dump(), **at.model_dump()}
        else:
            raise ValueError(
                f'conversation_type must be "1" or "2" not {conversation_type}'
            )

        try:
            async with self.session.post(webhook, json=data) as resp:
                return await resp.json()
        except aiohttp.ClientError as e:
            raise NetworkError from e



class libcore:
    """This is the core library of the package. It contains the main functions to be used by the user."""
    def sum_as_string(self, a: int, b: int) -> int:
        """This function returns the sum of two numbers."""
        ...



"""iamai message module.

Implemented the commonly used basic message ``Message`` and message field ``MessageSegment`` models for adapter use.
Adapter developers can implement subclasses of the message classes in this module or define different message types as needed, but it is recommended that subclasses of the message classes in this module be used if possible.
"""

from abc import ABC, abstractmethod
from typing import (
    Any,
    Dict,
    Generic,
    ItemsView,
    Iterator,
    KeysView,
    List,
    Mapping,
    Optional,
    SupportsIndex,
    Type,
    TypeVar,
    Union,
    ValuesView,
    overload,
)
from typing_extensions import Self

from pydantic import BaseModel, Field, GetCoreSchemaHandler
from pydantic_core import core_schema

__all__ = [
    "MessageT",
    "MessageSegmentT",
    "BuildMessageType",
    "Message",
    "MessageSegment",
]

MessageT = TypeVar("MessageT", bound="Message[Any]")
MessageSegmentT = TypeVar("MessageSegmentT", bound="MessageSegment[Any]")

# Type that can be converted to Message
BuildMessageType = Union[List[MessageSegmentT], MessageSegmentT, str, Mapping[str, Any]]


class Message(ABC, List[MessageSegmentT]):
    """Message.

    This class is a subclass of ``List`` and overrides the ``__init__()`` method.
    Can handle `str`, ``Mapping``, ``MessageSegment``, ``List[MessageSegment]`` directly.
    This class overloads the ``+`` and ``+=`` operators, and can directly perform sum operations with objects of ``Message``, ``MessageSegment`` and other types of objects.
    Adapter developers need to inherit this class and override the ``get_segment_class()`` method.
    """

    def __init__(self, *messages: BuildMessageType[MessageSegmentT]) -> None:
        """initialization.

        Args:
            *messages: Data that can be converted into messages.
        """
        segment_class = self.get_segment_class()
        for message in messages:
            if isinstance(message, list):
                self.extend(message)
            elif isinstance(message, segment_class):
                self.append(message)
            elif isinstance(message, str):
                self.append(segment_class.from_str(message))
            elif isinstance(message, Mapping):
                self.append(segment_class.from_mapping(message))
            else:
                raise TypeError(
                    f"message type error, expect List[{segment_class}], "
                    f"{segment_class}, str or Mapping, get {type(message)}"
                )

    @classmethod
    @abstractmethod
    def get_segment_class(cls) -> Type[MessageSegmentT]:
        """Get the message field class.

        Returns:
            Message field class.
        """

    @classmethod
    def __get_pydantic_core_schema__(
        cls, _source: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        """Pydantic custom mode."""
        return core_schema.union_schema(
            [
                core_schema.is_instance_schema(cls),
                core_schema.no_info_after_validator_function(
                    cls,
                    handler.generate_schema(
                        List[cls.get_segment_class()]  # type: ignore[misc, index]
                    ),
                ),
            ]
        )

    def __repr__(self) -> str:
        """Return the description of the message.

        Returns:
            Description of the message.
        """
        return f"Message:[{','.join(map(repr, self))}]"

    def __str__(self) -> str:
        """Returns the textual representation of the message.

        Returns:
            The textual representation of the message.
        """
        return "".join(map(str, self))

    def __contains__(self, item: object) -> bool:
        """Determine whether the message contains the specified text or message field.

        Args:
            item: text or message field.

        Returns:
            Whether the specified text or message field is included in the message.
        """
        if isinstance(item, str):
            return item in str(self)
        return super().__contains__(item)

    def __add__(self, other: BuildMessageType[MessageSegmentT]) -> Self:  # type: ignore
        """Method for adding custom messages to other objects.

        Args:
            other: other objects.

        Returns:
            The result of the addition.
        """
        return self.__class__(self).__iadd__(other)

    def __radd__(self, other: BuildMessageType[MessageSegmentT]) -> Self:
        """Method for adding custom messages to other objects.

        Args:
            other: other objects.

        Returns:
            The result of the addition.
        """
        return self.__class__(other).__iadd__(self)

    def __iadd__(self, other: BuildMessageType[MessageSegmentT]) -> Self:  # type: ignore
        """Method for adding custom messages to other objects.

        Args:
            other: other objects.

        Returns:
            The result of the addition.
        """
        try:
            self.extend(self.__class__(other))
        except TypeError as e:
            raise TypeError(
                f"unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'",
            ) from e
        return self

    def is_text(self) -> bool:
        """Whether it is a plain text message."""
        return all(x.is_text() for x in self)

    def get_plain_text(self) -> str:
        """Get the plain text part of the message.

        Returns:
            The plain text portion of the message.
        """
        return "".join(map(str, filter(lambda x: x.is_text(), self)))

    def copy(self) -> Self:
        """Return a shallow copy of itself.

        Returns:
            A shallow copy of itself.
        """
        return self.__class__(self)

    def startswith(
        self,
        prefix: Union[str, MessageSegmentT],
        start: Optional[SupportsIndex] = None,
        end: Optional[SupportsIndex] = None,
    ) -> bool:
        """Implement string-like ``startswith()`` method.

        When the ``prefix`` type is ``str``, it will convert itself to the ``str`` type, and then call the ``startswith()`` method of the ``str`` type.
        When the ``prefix`` type is ``MessageSegment``, the ``start`` and `end` parameters will have no effect,
            It will be judged whether the first message field of the list is equal to ``prefix``.

        Args:
            prefix: prefix.
            start: Start checking the position.
            end: Stop checking the position.

        Returns:
            test result.
        """  # noqa: D402
        if isinstance(prefix, str):
            return str(self).startswith(prefix, start, end)
        if isinstance(prefix, self.get_segment_class()):
            return False if len(self) == 0 else self[0] == prefix
        raise TypeError(
            f"first arg must be str or {self.get_segment_class()}, not {type(prefix)}"
        )

    def endswith(
        self,
        suffix: Union[str, MessageSegmentT],
        start: Optional[SupportsIndex] = None,
        end: Optional[SupportsIndex] = None,
    ) -> bool:
        """Implement string-like ``endswith()`` method.

        When the ``suffix`` type is `str`, it will convert itself to the ``str`` type, and then call the ``endswith()`` method of the ``str`` type.
        When the ``suffix`` type is MessageSegment, the `start` and ``end`` parameters will have no effect.
            Will determine whether the last message field in the list is equal to `suffix`.

        Args:
            suffix: suffix.
            start: Start checking the position.
            end: Stop checking the position.

        Returns:
            test result.
        """  # noqa: D402
        if isinstance(suffix, str):
            return str(self).endswith(suffix, start, end)
        if isinstance(suffix, self.get_segment_class()):
            return False if len(self) == 0 else self[-1] == suffix
        raise TypeError(
            f"first arg must be str or {self.get_segment_class()}, not {type(suffix)}"
        )

    @overload
    def replace(self, old: str, new: str, count: int = -1) -> Self: ...

    @overload
    def replace(
        self, old: MessageSegmentT, new: Optional[MessageSegmentT], count: int = -1
    ) -> Self: ...

    def replace(
        self,
        old: Union[str, MessageSegmentT],
        new: Optional[Union[str, MessageSegmentT]],
        count: int = -1,
    ) -> Self:
        """Implement string-like ``replace()`` method.

        When ``old`` is of type ``str``, ``new`` must also be ``str``, and this method will only process the message fields where ``is_text()`` is ``True``.
        When ``old`` is of MessageSegment type, ``new`` can be ``MessageSegment`` or ``None``, and this method will process all message fields.
            And replace the message fields that meet the criteria. ``None`` means to delete the matching message field.

        Args:
            old: The matched string or message field.
            new: The string or message field to be replaced.
            count: the number of replacements.

        Returns:
            The replaced message object.
        """  # noqa: D402
        if isinstance(old, str):
            if not isinstance(new, str):
                raise TypeError("when type of old is str, type of new must be str.")
            return self._replace_str(old, new, count)
        if isinstance(old, self.get_segment_class()):
            if not (isinstance(new, self.get_segment_class()) or new is None):
                raise TypeError(
                    "when type of old is MessageSegment, "
                    "type of new must be MessageSegment or None."
                )
            new_msg = self.__class__()
            for item in self:
                if count != 0 and item == old:
                    count -= 1
                    if new is not None:
                        new_msg.append(new)
                else:
                    new_msg.append(item)
            return new_msg
        raise TypeError("type of old must be str or MessageSegment")

    def _replace_str(self, old: str, new: str, count: int = -1) -> Self:
        """Implement string-like ``replace()`` method.

        This method will be called by the ``replace()`` method to handle replacement of type ``str``,
        By default, the ``data['text']`` of the ``MessageSegment`` object is treated as a location to store plain text.
        Adapter developers can override this method to adapt to other situations.

        Args:
            old: The matched string or message field.
            new: The string or message field to be replaced.
            count: the number of replacements.

        Returns:
            The replaced message object.
        """
        temp_msg = self.__class__(*(x.model_copy(deep=True) for x in self))
        for index, item in enumerate(temp_msg):
            if count == 0:
                break
            if item.is_text() and old in item.data["text"]:
                if count == -1:
                    temp_msg[index].data["text"] = item.data["text"].replace(old, new)
                else:
                    replace_times = min(count, item.data["text"].count(old))
                    temp_msg[index].data["text"] = item.data["text"].replace(
                        old, new, replace_times
                    )
                    count -= replace_times
        return temp_msg


class MessageSegment(ABC, BaseModel, Mapping[str, Any], Generic[MessageT]):
    """Message field.

    This class implements all ``Mapping`` type methods, all of which operate on the ``data`` attribute.
    This class overrides the ``+`` and ``+=`` operators, and can directly perform sum operations with objects of types such as ``Message``, ``MessageSegment`` and return a ``Message`` object.
    Adapter developers need to inherit this class and override the ``get_message_class()`` method.

    Attributes:
        type: message field type.
        data: message field content.
    """

    type: str
    data: Dict[str, Any] = Field(default_factory=dict)

    @classmethod
    @abstractmethod
    def get_message_class(cls) -> Type[MessageT]:
        """Get the message class.

        Returns:
            Message class.
        """

    @classmethod
    @abstractmethod
    def from_str(cls, msg: str) -> Self:
        """Used to convert ``str`` to a message field, subclasses should override this method.

        Args:
            msg: Data to be parsed into message fields.

        Returns:
            Message fields converted by ``str``.
        """

    @classmethod
    def from_mapping(cls, msg: Mapping[Any, Any]) -> Self:
        """Used to convert ``Mapping`` into message fields.

        Subclasses can override this method to change the default behavior for ``Mapping`` if necessary.

        Args:
            msg: Data to be parsed into message fields.

        Returns:
            Message fields converted by Mapping.
        """
        return cls(**msg)

    def __str__(self) -> str:
        """Returns the text representation of the message field.

        Returns:
            Text representation of the message field.
        """
        return str(self.data)

    def __repr__(self) -> str:
        """Returns the description of the message field.

        Returns:
            Description of message fields.
        """
        return f"MessageSegment<{self.type}>:{self!s}"

    def __getitem__(self, key: str) -> Any:
        """Get the index. It is equivalent to doing this operation on the ``data`` attribute.

        Args:
            key: key.

        Returns:
            `data` The value of the corresponding index in the dictionary.
        """
        return self.data[key]

    def __setitem__(self, key: str, value: Any) -> None:
        """Set the value of the specified index. Equivalent to doing this on the ``data`` attribute.

        Args:
            key: key.
            value: value.
        """
        self.data[key] = value

    def __delitem__(self, key: str) -> None:
        """Delete the index. Equivalent to doing this operation on the ``data`` attribute.

        Args:
            key: key.
        """
        del self.data[key]

    def __len__(self) -> int:
        """Get the length. Equivalent to doing this operation on the ``data` attribute.

        Returns:
            `data` The length of the dictionary.
        """
        return len(self.data)

    def __iter__(self) -> Iterator[str]:  # type: ignore
        """Iterate. Equivalent to doing this operation on the ``data`` attribute.

        Returns:
            ``data`` Iterator for dictionary.
        """
        yield from self.data.__iter__()

    def __contains__(self, key: object) -> bool:
        """Whether the index is included in the object. Equivalent to doing this on the ``data`` attribute.

        Args:
            key: key.

        Returns:
            Whether the index is contained in the ``data`` dictionary.
        """
        return key in self.data

    def __eq__(self, other: object) -> bool:
        """Determine whether they are equal.

        Args:
            other: other objects.

        Returns:
            are equal.
        """
        return (
            type(other) is self.__class__  # pylint: disable=unidiomatic-typecheck
            and self.type == other.type
            and self.data == other.data
        )

    def __ne__(self, other: object) -> bool:
        """Determine whether they are not equal.

        Args:
            other: other objects.

        Returns:
            Whether they are not equal.
        """
        return not self.__eq__(other)

    def __add__(self, other: Any) -> MessageT:
        """Method for adding custom message fields to other objects.

        Args:
            other: other objects.

        Returns:
            The result of the addition.
        """
        return self.get_message_class()(self) + other

    def __radd__(self, other: Any) -> MessageT:
        """Method for adding custom message fields to other objects.

        Args:
            other: other objects.

        Returns:
            The result of the addition.
        """
        return self.get_message_class()(other) + self

    def get(self, key: str, default: Any = None) -> Any:
        """Returns the value of ``key`` if ``key`` exists in the ``data`` dictionary, otherwise returns ``default``."""
        return self.data.get(key, default)

    def keys(self) -> KeysView[str]:
        """Returns a new view consisting of ``data`` dictionary keys."""
        return self.data.keys()

    def values(self) -> ValuesView[Any]:
        """Returns a new view consisting of ``data`` dictionary values."""
        return self.data.values()

    def items(self) -> ItemsView[str, Any]:
        """Returns a new view consisting of ``data`` dictionary items (``(key, value)`` pairs)."""
        return self.data.items()

    def is_text(self) -> bool:
        """is a plain text message field.

        Returns:
            Whether it is a plain text message field.
        """
        return self.type == "text"



"""Console 适配器事件。"""

from typing import (
    TYPE_CHECKING,
    TypeVar,
)

from pydantic import BaseModel

from iamai.event import Event

from .message import Message

T_ConsoleEvent = TypeVar("T_ConsoleEvent", bound="ConsoleEvent")

if TYPE_CHECKING:
    pass  # type: ignore[class]

__all__ = ["ConsoleEvent", "MessageEvent", "User", "Robot"]  # noqa: F822


class User(BaseModel, frozen=True):
    """用户"""

    id: str
    avatar: str = "👤"
    nickname: str = "User"


class Robot(User, frozen=True):
    """机器人"""

    avatar: str = "🤖"
    nickname: str = "Bot"


class ConsoleEvent(Event["ConsoleAdapter"]):
    """Console 事件基类。"""

    message: str

    def get_event_description(self) -> str:
        return str(self.dict())

    def get_message(self) -> Message:
        raise ValueError("Event has no message!")

    def get_user_id(self) -> str:
        raise ValueError("Event has no user_id!")

    def get_session_id(self) -> str:
        raise ValueError("Event has no session_id!")

    def is_tome(self) -> bool:
        """获取事件是否与机器人有关的方法。"""
        return True



"""APScheduler 适配器事件。"""

from typing import TYPE_CHECKING, Any, Dict, Optional, Type, Union

from apscheduler.job import Job
from apscheduler.triggers.base import BaseTrigger

from iamai.event import Event
from iamai.plugin import Plugin

if TYPE_CHECKING:
    pass


__all__ = ["APSchedulerEvent"]


class APSchedulerEvent(Event["APSchedulerAdapter"]):
    """APSchedulerEvent 事件基类。"""

    type: Optional[str] = "apscheduler"
    plugin_class: Type[Plugin]  # type: ignore

    @property
    def job(self) -> Job:
        """产生当前事件的 APScheduler `Job` 对象。"""
        return self.adapter.plugin_class_to_job[self.plugin_class]

    @property
    def trigger(self) -> Union[str, BaseTrigger]:
        """当前事件对应的 Plugin 的 `trigger`。"""
        return getattr(self.plugin_class, "trigger")  # noqa: B009

    @property
    def trigger_args(self) -> Dict[str, Any]:
        """当前事件对应的 Plugin 的 `trigger_args`。"""
        return getattr(self.plugin_class, "trigger_args")  # noqa: B009



from iamai import Plugin


class Count(Plugin):
    async def handle(self) -> None:
        if self.state is None:
            self.state = 0
        self.state += 1
        await self.event.reply(f"count: {self.state}")

    async def rule(self) -> bool:
        if self.event.adapter.name != "cqhttp":
            return False
        if self.event.type != "message_sent":
            return False
        return self.event.message.get_plain_text() == "count"



import re

from iamai.message import MessageSegment


class ConsoleMessage(MessageSegment[None]):
    """Console 适配器消息。"""

    @property
    def _message_class(self) -> None:
        return None

    def is_text(self) -> bool:
        return self.type == "text"


def escape_tag(s: str) -> str:
    """用于记录带颜色日志时转义 `<tag>` 类型特殊标签

    参考: [loguru color 标签](https://loguru.readthedocs.io/en/stable/api/logger.html#color)

    参数:
        s: 需要转义的字符串
    """
    return re.sub(r"</?((?:[fb]g\s)?[^<>\s]*)>", r"\\\g<0>", s)



"""CQHTTP 适配器事件。"""

# pyright: reportIncompatibleVariableOverride=false

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Literal,
    Optional,
    Tuple,
    get_args,
    get_origin,
)
from typing_extensions import Self

from pydantic import BaseModel, ConfigDict, Field
from pydantic.fields import FieldInfo

from iamai.event import Event
from iamai.event import MessageEvent as BaseMessageEvent
from iamai.message import BuildMessageType

from .message import CQHTTPMessage, CQHTTPMessageSegment

if TYPE_CHECKING:
    pass


class Sender(BaseModel):
    """发送人信息"""

    user_id: Optional[int] = None
    nickname: Optional[str] = None
    card: Optional[str] = None
    sex: Optional[Literal["male", "female", "unknown"]] = None
    age: Optional[int] = None
    area: Optional[str] = None
    level: Optional[str] = None
    role: Optional[str] = None
    title: Optional[str] = None


class Anonymous(BaseModel):
    """匿名信息"""

    id: int
    name: str
    flag: str


class File(BaseModel):
    """文件信息"""

    id: str
    name: str
    size: int
    busid: int


class Status(BaseModel):
    """状态信息"""

    model_config = ConfigDict(extra="allow")

    online: bool
    good: bool


def _get_literal_field(field: Optional[FieldInfo]) -> Optional[str]:
    if field is None:
        return None
    annotation = field.annotation
    if annotation is None or get_origin(annotation) is not Literal:
        return None
    literal_values = get_args(annotation)
    if len(literal_values) != 1:
        return None
    return literal_values[0]


class CQHTTPEvent(Event["CQHTTPAdapter"]):
    """CQHTTP 事件基类"""

    __event__ = ""
    type: Optional[str] = Field(alias="post_type")
    time: int
    self_id: int
    post_type: str

    @property
    def to_me(self) -> bool:
        """当前事件的 `user_id` 是否等于 `self_id`。"""
        return getattr(self, "user_id", None) == self.self_id

    @classmethod
    def get_event_type(cls) -> Tuple[Optional[str], Optional[str], Optional[str]]:
        """获取事件类型。

        Returns:
            事件类型。
        """
        post_type = _get_literal_field(cls.model_fields.get("post_type", None))
        if post_type is None:
            return (None, None, None)
        return (
            post_type,
            _get_literal_field(cls.model_fields.get(post_type + "_type", None)),
            _get_literal_field(cls.model_fields.get("sub_type", None)),
        )


class MessageEvent(CQHTTPEvent, BaseMessageEvent["CQHTTPAdapter"]):
    """消息事件"""

    __event__ = "message"
    post_type: Literal["message"]
    message_type: Literal["private", "group"]
    sub_type: str
    message_id: int
    user_id: int
    message: CQHTTPMessage
    raw_message: str
    font: int
    sender: Sender

    def __repr__(self) -> str:
        """返回消息事件的描述。

        Returns:
            消息事件的描述。
        """
        return f'Event<{self.type}>: "{self.message}"'

    def get_plain_text(self) -> str:
        """获取消息的纯文本内容。

        Returns:
            消息的纯文本内容。
        """
        return self.message.get_plain_text()

    async def reply(
        self, message: BuildMessageType[CQHTTPMessageSegment]
    ) -> Dict[str, Any]:
        """回复消息。

        Args:
            message: 回复消息的内容，同 `call_api()` 方法。

        Returns:
            API 请求响应。
        """
        raise NotImplementedError

    async def is_same_sender(self, other: Self) -> bool:
        """判断自身和另一个事件是否是同一个发送者。

        Args:
            other: 另一个事件。

        Returns:
            是否是同一个发送者。
        """
        return self.sender.user_id == other.sender.user_id


class PrivateMessageEvent(MessageEvent):
    """私聊消息"""

    __event__ = "message.private"
    message_type: Literal["private"]
    sub_type: Literal["friend", "group", "other"]

    async def reply(
        self, message: BuildMessageType[CQHTTPMessageSegment]
    ) -> Dict[str, Any]:
        """回复消息。

        Args:
            message: 回复消息的内容，同 `call_api()` 方法。

        Returns:
            API 请求响应。
        """
        return await self.adapter.send_private_msg(
            user_id=self.user_id, message=CQHTTPMessage(message)
        )


class GroupMessageEvent(MessageEvent):
    """群消息"""

    __event__ = "message.group"
    message_type: Literal["group"]
    sub_type: Literal["normal", "anonymous", "notice"]
    group_id: int
    anonymous: Optional[Anonymous] = None

    async def reply(
        self, message: BuildMessageType[CQHTTPMessageSegment]
    ) -> Dict[str, Any]:
        """回复消息。

        Args:
            message: 回复消息的内容，同 `call_api()` 方法。

        Returns:
            API 请求响应。
        """
        return await self.adapter.send_group_msg(
            group_id=self.group_id, message=CQHTTPMessage(message)
        )


class NoticeEvent(CQHTTPEvent):
    """通知事件"""

    __event__ = "notice"
    post_type: Literal["notice"]
    notice_type: str


class GroupUploadNoticeEvent(NoticeEvent):
    """群文件上传"""

    __event__ = "notice.group_upload"
    notice_type: Literal["group_upload"]
    user_id: int
    group_id: int
    file: File


class GroupAdminNoticeEvent(NoticeEvent):
    """群管理员变动"""

    __event__ = "notice.group_admin"
    notice_type: Literal["group_admin"]
    sub_type: Literal["set", "unset"]
    user_id: int
    group_id: int


class GroupDecreaseNoticeEvent(NoticeEvent):
    """群成员减少"""

    __event__ = "notice.group_decrease"
    notice_type: Literal["group_decrease"]
    sub_type: Literal["leave", "kick", "kick_me"]
    group_id: int
    operator_id: int
    user_id: int


class GroupIncreaseNoticeEvent(NoticeEvent):
    """群成员增加"""

    __event__ = "notice.group_increase"
    notice_type: Literal["group_increase"]
    sub_type: Literal["approve", "invite"]
    group_id: int
    operator_id: int
    user_id: int


class GroupBanNoticeEvent(NoticeEvent):
    """群禁言"""

    __event__ = "notice.group_ban"
    notice_type: Literal["group_ban"]
    sub_type: Literal["ban", "lift_ban"]
    group_id: int
    operator_id: int
    user_id: int
    duration: int


class FriendAddNoticeEvent(NoticeEvent):
    """好友添加"""

    __event__ = "notice.friend_add"
    notice_type: Literal["friend_add"]
    user_id: int


class GroupRecallNoticeEvent(NoticeEvent):
    """群消息撤回"""

    __event__ = "notice.group_recall"
    notice_type: Literal["group_recall"]
    group_id: int
    operator_id: int
    user_id: int
    message_id: int


class FriendRecallNoticeEvent(NoticeEvent):
    """好友消息撤回"""

    __event__ = "notice.friend_recall"
    notice_type: Literal["friend_recall"]
    user_id: int
    message_id: int


class NotifyEvent(NoticeEvent):
    """提醒事件"""

    __event__ = "notice.notify"
    notice_type: Literal["notify"]
    sub_type: str
    group_id: Optional[int] = None
    user_id: int


class PokeNotifyEvent(NotifyEvent):
    """戳一戳"""

    __event__ = "notice.notify.poke"
    sub_type: Literal["poke"]
    target_id: int
    group_id: Optional[int] = None


class GroupLuckyKingNotifyEvent(NotifyEvent):
    """群红包运气王"""

    __event__ = "notice.notify.lucky_king"
    sub_type: Literal["lucky_king"]
    group_id: int
    target_id: int


class GroupHonorNotifyEvent(NotifyEvent):
    """群成员荣誉变更"""

    __event__ = "notice.notify.honor"
    sub_type: Literal["honor"]
    group_id: int
    honor_type: Literal["talkative", "performer", "emotion"]


class RequestEvent(CQHTTPEvent):
    """请求事件"""

    __event__ = "request"
    post_type: Literal["request"]
    request_type: str

    async def approve(self) -> Dict[str, Any]:
        """同意请求。

        Returns:
            API 请求响应。
        """
        raise NotImplementedError

    async def refuse(self) -> Dict[str, Any]:
        """拒绝请求。

        Returns:
            API 请求响应。
        """
        raise NotImplementedError


class FriendRequestEvent(RequestEvent):
    """加好友请求"""

    __event__ = "request.friend"
    request_type: Literal["friend"]
    user_id: int
    comment: str
    flag: str

    async def approve(self, remark: str = "") -> Dict[str, Any]:
        """同意请求。

        Args:
            remark: 好友备注。

        Returns:
            API 请求响应。
        """
        return await self.adapter.set_friend_add_request(
            flag=self.flag, approve=True, remark=remark
        )

    async def refuse(self) -> Dict[str, Any]:
        """拒绝请求。

        Returns:
            API 请求响应。
        """
        return await self.adapter.set_friend_add_request(flag=self.flag, approve=False)


class GroupRequestEvent(RequestEvent):
    """加群请求 / 邀请"""

    __event__ = "request.group"
    request_type: Literal["group"]
    sub_type: Literal["add", "invite"]
    group_id: int
    user_id: int
    comment: str
    flag: str

    async def approve(self) -> Dict[str, Any]:
        """同意请求。

        Returns:
            API 请求响应。
        """
        return await self.adapter.set_group_add_request(
            flag=self.flag, sub_type=self.sub_type, approve=True
        )

    async def refuse(self, reason: str = "") -> Dict[str, Any]:
        """拒绝请求。

        Args:
            reason: 拒绝原因。

        Returns:
            API 请求响应。
        """
        return await self.adapter.set_group_add_request(
            flag=self.flag, sub_type=self.sub_type, approve=False, reason=reason
        )


class MetaEvent(CQHTTPEvent):
    """元事件"""

    __event__ = "meta_event"
    post_type: Literal["meta_event"]
    meta_event_type: str


class LifecycleMetaEvent(MetaEvent):
    """生命周期"""

    __event__ = "meta_event.lifecycle"
    meta_event_type: Literal["lifecycle"]
    sub_type: Literal["enable", "disable", "connect"]


class HeartbeatMetaEvent(MetaEvent):
    """心跳"""

    __event__ = "meta_event.heartbeat"
    meta_event_type: Literal["heartbeat"]
    status: Status
    interval: int



import os
import re
import json
import pathlib

from python_graphql_client import GraphqlClient

root = pathlib.Path(__file__).parent.resolve()
client = GraphqlClient(endpoint="https://api.github.com/graphql")

with open(f"{root}/credit.json", "r", encoding="utf-8") as credit_file:
    credit = json.loads(credit_file.read())

TOKEN = os.environ.get("GITHUB_TOKEN", "")


def replace_chunk(content, marker, chunk, inline=False):
    r = re.compile(
        f"<!\-\- {marker} starts \-\->.*<!\-\- {marker} ends \-\->", re.DOTALL
    )
    if not inline:
        chunk = f"\n{chunk}\n"
    chunk = f"<!-- {marker} starts -->{chunk}<!-- {marker} ends -->"
    return r.sub(chunk, content)


def fetch_mlf_denpendices():
    dependencies = credit["MachineLearningFramework"]
    return [
        {
            "name": _["name"],
            "version": _["versionInfo"],
            "license": _["license"],
            "repo": _["repo"],
            "licenseUrl": _["licenseUrl"],
            "description": _["description"],
            "icon": _["icon"],
        }
        for _ in dependencies
    ]


def fetch_cprf_denpendices():
    dependencies = credit["RobotFramework"]
    return [
        {
            "name": _["name"],
            "version": _["versionInfo"],
            "license": _["license"],
            "repo": _["repo"],
            "licenseUrl": _["licenseUrl"],
            "description": _["description"],
            "icon": _["icon"],
        }
        for _ in dependencies
    ]


if __name__ == "__main__":
    readme = root / "credits.md"
    readme_contents = readme.open().read()
    # MLF
    MLF = fetch_mlf_denpendices()
    MLF_md = "\n\n".join(
        [
            "[{name}]({repo})({version}) with [{license}]({licenseUrl}). {icon} `{description}`".format(
                **_
            )
            for _ in MLF
        ]
    )
    print()
    print(MLF_md)
    print()
    rewritten = replace_chunk(readme_contents, "MLF", MLF_md)

    # CPRF
    CPRF = fetch_cprf_denpendices()
    CPRF_md = "\n\n".join(
        [
            "[{name}]({repo})({version}) with [{license}]({licenseUrl}). {icon} `{description}`".format(
                **_
            )
            for _ in CPRF
        ]
    )
    print()
    print(CPRF_md)
    print()
    rewritten = replace_chunk(rewritten, "CPRF", CPRF_md)

    readme.open("w").write(rewritten)



from time import strftime, localtime

from iamai import Plugin
from iamai.adapter.apscheduler import scheduler_decorator


@scheduler_decorator(
    trigger="interval", trigger_args={"seconds": 10}, override_rule=True
)
class Schedulers(Plugin):
    async def handle(self) -> None:
        await self.bot.get_adapter("console").send(
            f"Time: {strftime('%Y-%m-%d %H:%M:%S', localtime())}"
        )

    async def rule(self) -> bool:
        return False



import os
import time
import asyncio
import logging
from typing import Any

import qrcode
import aiohttp
from aiohttp import cookiejar
from aiohttp.client import ClientSession
from aiohttp.client_exceptions import ClientResponseError

QRCODE_REQUEST_URL = "http://passport.bilibili.com/qrcode/getLoginUrl"
CHECK_LOGIN_RESULT = "http://passport.bilibili.com/qrcode/getLoginInfo"
SEND_URL = "https://api.live.bilibili.com/msg/send"
MUTE_USER_URL = (
    "https://api.live.bilibili.com/xlive/web-ucenter/v1/banned/AddSilentUser"
)
ROOM_SLIENT_URL = "https://api.live.bilibili.com/xlive/web-room/v1/banned/RoomSilent"
ADD_BADWORD_URL = (
    "https://api.live.bilibili.com/xlive/web-ucenter/v1/banned/AddShieldKeyword"
)
DEL_BADWORD_URL = (
    "https://api.live.bilibili.com/xlive/web-ucenter/v1/banned/DelShieldKeyword"
)


user_cookies = cookiejar.CookieJar()

"""
Bilibili Client Operation

"""


async def login(session: ClientSession) -> bool:
    if get_cookies("bili_jct") != None:  # noqa: E711
        # 無需重複獲取
        logging.info("先前已經登入，因此無需再度登入。")
        return True
    try:
        res = await _get(session, QRCODE_REQUEST_URL)

        ts = res["ts"]
        outdated = ts + 180 * 1000  # 180 秒後逾時
        authKey = res["data"]["oauthKey"]

        url = res["data"]["url"]
        qr = qrcode.QRCode()
        logging.info("請掃描下列二維碼進行登入... (或者到目錄下尋找 qrcode.png)")

        qr.add_data(url)
        qr.print_ascii(invert=True)
        qr.make_image().save("qrcode.png")

        while True:
            await asyncio.sleep(5)

            if time.time() > outdated:
                logging.info("已逾時。")
                return False  # 登入失敗

            res = await _post(session, CHECK_LOGIN_RESULT, oauthKey=authKey)

            if res["status"]:
                logging.info("登入成功。")
                return True
            else:
                code = res["data"]
                if code in [-1, -2]:
                    logging.warning(f'登入失敗: {res["message"]}')
                    return False

    except ClientResponseError as e:
        logging.warning(f"請求時出現錯誤: {e}")
        return False
    finally:
        os.remove("qrcode.png")


async def send_danmu(**fields) -> bool:
    token = get_cookies("bili_jct")
    async with ClientSession(cookie_jar=user_cookies) as session:
        try:
            res = await _post(
                session,
                SEND_URL,
                rnd=time.time(),
                csrf=token,
                csrf_token=token,
                **fields,
            )
            return "data" in res
        except Exception as e:
            logging.warning(f"發送彈幕時出現錯誤: {e}")
            return False


def get_cookies(name: str) -> Any:
    for cookie in user_cookies:
        if cookie.key == name:
            return cookie.value
    return None


async def mute_user(tuid: int, roomid: int) -> bool:
    token = get_cookies("bili_jct")
    async with ClientSession(cookie_jar=user_cookies) as session:
        try:
            res = await _post(
                session,
                MUTE_USER_URL,
                csrf=token,
                csrf_token=token,
                visit_id="",
                mobile_app="web",
                tuid=str(tuid),
                room_id=str(roomid),
            )
            return res["code"] == 0
        except Exception as e:
            logging.warning(f"禁言時出現錯誤: {e}")
            return False


async def room_slient(roomid: int, slientType: str, level: int, minute: int) -> bool:
    type_availables = ["off", "medal", "member", "level"]
    if slientType not in type_availables:
        logging.warning(f"未知的禁言類型: {slientType} ({type_availables})")
        return False

    minute_available = [0, 30, 60]
    if minute not in minute_available:
        logging.warning(f"未知的静音时间: {minute} ({minute_available})")
        return False

    token = get_cookies("bili_jct")
    async with ClientSession(cookie_jar=user_cookies) as session:
        try:
            res = await _post(
                session,
                ROOM_SLIENT_URL,
                csrf=token,
                csrf_token=token,
                visit_id="",
                room_id=str(roomid),
                type=str(slientType),
                minute=str(minute),
                level=str(level),
            )
            return res["code"] == 0
        except Exception as e:
            logging.warning(f"房間靜音時出現錯誤: {e}")
            return False


async def add_badword(roomid: int, keyword: str) -> bool:
    token = get_cookies("bili_jct")
    async with ClientSession(cookie_jar=user_cookies) as session:
        try:
            res = await _post(
                session,
                ADD_BADWORD_URL,
                csrf=token,
                csrf_token=token,
                visit_id="",
                room_id=str(roomid),
                keyword=keyword,
            )
            return res["code"] == 0
        except Exception as e:
            logging.warning(f"添加屏蔽字時出現錯誤: {e}")
            return False


async def remove_badword(roomid: int, keyword: str) -> bool:
    token = get_cookies("bili_jct")
    async with ClientSession(cookie_jar=user_cookies) as session:
        try:
            res = await _post(
                session,
                DEL_BADWORD_URL,
                csrf=token,
                csrf_token=token,
                visit_id="",
                room_id=str(roomid),
                keyword=keyword,
            )
            return res["code"] == 0
        except Exception as e:
            logging.warning(f"删除屏蔽字時出現錯誤: {e}")
            return False


def logout():
    user_cookies.clear()


"""
Http Request

"""


async def _get(session: ClientSession, url: str):
    async with session.get(url) as resp:
        resp.raise_for_status()
        data = await resp.json()
        logging.debug(data)
        if "code" in data and data["code"] != 0:
            raise Exception(data["message"] if "message" in data else data["code"])
        return data


async def _post(session: ClientSession, url: str, **data):
    form = aiohttp.FormData()
    for k, v in data.items():
        form.add_field(k, v)
    logging.debug(f"正在发送 POST 请求: {url}, 内容: {data}")
    async with session.post(url, data=form) as resp:
        resp.raise_for_status()
        data = await resp.json()
        logging.debug(data)
        if "code" in data and data["code"] != 0:
            raise Exception(data["message"] if "message" in data else data["code"])
        return data


if __name__ == "__main__":
    session = ClientSession(cookies={"a": 1, "b": 2})
    for c in session.cookie_jar:
        print(c.key, c.value)



"""Kook Adapter

This adapter is adapted to the Kook Platform.
For details of the agreement, please refer to: [Kook Developer Platform](https://developer.kookapp.cn/)
"""

import re
import sys  # noqa: F401
import json
import time  # noqa: F401
import zlib
import asyncio
from functools import partial
from typing import TYPE_CHECKING, Any, Dict, Literal, Mapping, Optional

import aiohttp
import requests
from pydantic import parse_obj_as

from iamai.adapter.utils import WebSocketAdapter
from iamai.log import logger, error_or_exception

from .config import Config
from .message import MessageDeserializer, rev_msg_type_map  # noqa: F401
from .api.handle import User, get_api_method, get_api_restype
from .exceptions import (
    ApiTimeout,  # noqa: F401
    TokenError,
    ActionFailed,
    NetworkError,  # noqa: F401
    ReconnectError,
    ApiNotAvailable,  # noqa: F401
)
from .event import (
    KookEvent,
    EventTypes,  # noqa: F401
    OriginEvent,  # noqa: F401
    ResultStore,
    SignalTypes,
    _kook_events,  # noqa: F401
    get_event_class,  # noqa: F401
)

if TYPE_CHECKING:
    from .message import T_KookMSG

__all__ = ["KookAdapter"]

BASE_URL = "https://www.kookapp.cn/api"


class KookAdapter(WebSocketAdapter[KookEvent, Config]):
    """Kook Adapter."""

    name: str = "kook"
    Config = Config

    _gateway_response: dict = {}
    _api_response: Dict[Any, Any]
    _api_response_cond: asyncio.Condition
    _api_id: int = 0

    def __getattr__(self, item):
        return partial(self.call_api, item)

    def get_api_protocol(self, version_number: int | str = 3) -> str:
        """API version management
        KOOK may have different versions of API in the future. You can pass it like ``https://www.kookapp.cn/api/v{version_number}``
        This explicitly specifies the API version to use in the request path. If version_number is omitted, it will point to the default version.

            Specific reference: https://developer.kookapp.cn/doc/reference

        Args:
            version_number (int, optional): version code. Defaults to 3.

        Returns:
            str: KOOK API URL of the corresponding version
        """
        return f"{BASE_URL}/v{version_number}"

    def build_url(self, args) -> str:
        return "/".join(args)

    async def startup(self):
        """Initialize the adapter."""
        self.adapter_type = self.config.adapter_type
        if self.adapter_type == "websocket":
            self.adapter_type = "ws"
        self.bot.global_state["adapter"] = self.bot.global_state.get("adapter", {})
        self.bot.global_state["adapter"]["kook"] = {}
        self.reconnect_interval = self.config.reconnect_interval
        self._api_response_cond = asyncio.Condition()
        await super().startup()

    async def websocket_connect(self) -> None:
        """Create a forward WebSocket connection."""
        logger.info("Trying to GET the GateWay...")
        url = self.build_url([self.get_api_protocol(), "gateway", "index"])
        headers = {
            "Authorization": f"Bot {self.config.access_token}",
        }
        # Get The Gateway URL
        # https://developer.kookapp.cn/doc/http/gateway
        response = requests.get(
            url, headers=headers, params={"compress": self.config.compress}
        )
        if response.status_code == 200:
            logger.success("Successed to get GateWay.")
            self._gateway_response = response.json()
            self.bot.global_state["adapter"]["kook"][
                "bot_info"
            ] = await self._get_self_data(self.config.access_token)
            self.self_id = self.bot.global_state["adapter"]["kook"]["bot_info"].id_
            self.self_name = self.bot.global_state["adapter"]["kook"][
                "bot_info"
            ].username
            logger.success(f"Bot<{self.self_name}> self id: {self.self_id}")
        else:
            logger.error(f"Failed to get GateWay, status_code: {response.status_code}")
            return

        logger.info("Trying to connect to WebSocket server...")

        # start connection
        async with self.session.ws_connect(
            self._gateway_response["data"]["url"]
        ) as self.websocket:
            await self.handle_websocket()

    async def handle_websocket_msg(self, msg: aiohttp.WSMessage):
        """Handle Websocket Message."""
        msg_dict: dict
        if msg.type == aiohttp.WSMsgType.TEXT:
            try:
                msg_dict = msg.json()
                logger.debug(msg_dict)
            except json.JSONDecodeError as e:
                error_or_exception(
                    "WebSocket message parsing error, not json:",
                    e,
                    self.bot.config.bot.log.verbose_exception,
                )
                return

        elif msg.type == aiohttp.WSMsgType.BINARY:
            try:
                msg_dict: dict = zlib.decompress(msg.data).decode("utf-8")  # type: ignore[dict]
                logger.debug(msg_dict)
            except zlib.error as e:
                error_or_exception(
                    "WebSocket message decoding error, not binary:",
                    e,
                    self.bot.config.bot.log.verbose_exception,
                )
                return

        elif msg.type == aiohttp.WSMsgType.ERROR:
            logger.error(
                f"WebSocket connection closed "
                f"with exception {self.websocket.exception()!r}"  # type: ignore
            )
            return
        else:
            return

        # reveive hello package
        if msg_dict.get("s") == SignalTypes.HELLO:
            data: dict = msg_dict["d"]
            if data.get("code") == 0:
                try:
                    logger.success(
                        f"WebSocket connection verified, "
                        f"Session key: {data['session_id'][:7]}"
                    )
                    # Call start_heartbeat to send heartbeats at intervals of 30 (+5,-5)
                    self.bot.global_state["adapter"]["kook"]["session"] = data.get(
                        "session_id"
                    )
                    ResultStore.set_sn(
                        self.bot.global_state["adapter"]["kook"]["session"], 0
                    )
                    asyncio.ensure_future(
                        self.start_heartbeat(
                            self.bot.global_state["adapter"]["kook"]["session"]
                        )
                    )
                    logger.debug("HeartBeat task started!")
                    await self.handle_kook_event(data)
                except Exception as e:
                    logger.error(f"WebSocket connection verified failed!\n{e}")
                    raise ReconnectError from e
            elif data.get("code") == 40103:
                raise ReconnectError
            elif data.get("code") == 40101:
                raise TokenError("Invalid Token!")
            elif data.get("code") == 40102:
                raise TokenError("Token verification failed")
            else:
                logger.warning(
                    f"Websocket connection failed with code {msg_dict['d'].get('code') or msg_dict}, "
                    f"retrying..."
                )
                await asyncio.sleep(self.reconnect_interval)
        elif msg_dict.get("s") == SignalTypes.PONG:
            data = {
                "self_id": self.self_id,
                "post_type": "meta_event",
                "meta_event_type": "heartbeat",
            }
            logger.info(f"HeartBeat received!{data}")
            logger.info(
                f"Bot {self.bot.global_state['kook']['bot_info'].username} HeartBeat",
            )
            await self.handle_kook_event(data)
        elif msg_dict.get("s") == SignalTypes.EVENT:
            ResultStore.set_sn(self.bot.global_state["kook"]["session"], msg_dict["sn"])
            try:
                data = msg_dict["d"]
                await self.handle_kook_event(data)
            except Exception as e:
                logger.error(f"Event handle failed!\n{e!r}")
        elif msg_dict.get("s") == SignalTypes.RECONNECT:
            raise ReconnectError
        elif msg_dict.get("s") == SignalTypes.RESUME_ACK:
            return
        else:
            async with self._api_response_cond:
                self._api_response = msg_dict
                self._api_response_cond.notify_all()

    async def handle_kook_event(self, data: Dict[str, Any]):
        """Handle kook events.

        Args:
            msg: received message.
        """
        post_type = data.get("type")  # noqa: F841

        kook_event = KookEvent(adapter=self, **data)

        if self.config.show_raw:
            logger.debug(data)

        if kook_event.post_type == "meta_event":
            if (
                kook_event.meta_event_type == "lifecycle"
                and kook_event.sub_type == "connect"
            ):
                logger.success(
                    f"WebSocket connection "
                    f"from Kook Bot {self.bot.global_state['kook']['bot_info'].username} accepted!"
                )
        else:
            if (
                not self.config.report_self_message
                and kook_event.user_id == kook_event.self_id
            ):
                return
            await self.handle_event(kook_event)

    async def call_api(self, api: str, **data: dict) -> Any:
        match = re.findall(r"[A-Z]", api)
        if len(match) > 0:
            for m in match:
                api = api.replace(m, f"-{m.lower()}")
        api = api.replace("_", "/")

        if api.startswith("/api/v3/"):
            api = api[len("/api/v3/") :]
        elif api.startswith("api/v3"):
            api = api[len("api/v3") :]
        api = api.strip("/")
        return await self._call_api(api, data, self.config.access_token)  # type: ignore

    async def _call_api(
        self,
        api: str,
        data: Optional[Mapping[str, Any]] = None,
        token: Optional[str] = None,
    ) -> Any:
        data = dict(data) if data is not None else {}

        method = data.get("method") if data.get("method") else get_api_method(api)
        headers = data.get("headers", {})

        files = None
        query = None
        body = None

        if "files" in data:
            files = data["files"]
            del data["files"]
        elif "file" in data:
            files = {"file": data["file"]}
            del data["file"]

        if method == "GET":
            query = data
        elif method == "POST":
            body = data

        if token is not None:
            headers["Authorization"] = f"Bot {self.config.access_token}"

        result_type = get_api_restype(api)
        try:
            resp = requests.request(
                method=method,
                url=self.build_url([self.get_api_protocol(), api]),
                headers=headers,
                params=query,
                data=body,
                files=files,
                timeout=self.config.api_timeout,
            )
            result = _handle_api_result(resp)
            logger.debug(f"API {api} called with result {result}")
            return parse_obj_as(result_type, result) if result_type else None
        except Exception as e:
            raise e

    async def _get_self_data(self, token: str) -> User:
        """获取当前机器人的信息。

        Returns:
            Optional[dict]: 当前机器人的信息。
        """
        token = token or self.config.access_token
        return await self._call_api("user/me", token=token)

    async def start_heartbeat(self, session: str) -> None:
        """
        每30s一次心跳
        :return:
        """
        while not self.bot.should_exit.is_set() and not self.websocket.closed:
            await self.websocket.send_json(
                json.dumps({"s": 2, "sn": ResultStore.get_sn(session)})
            )
            logger.debug(f"HeartBeat sent {ResultStore.get_sn(session)} times!")
            await asyncio.sleep(26)

    async def send(
        self, message_: "T_KookMSG", message_type: Literal["GROUP", "PERSON"], id_: int
    ) -> Dict[str, Any]:
        """发送消息，调用 message/create 或 direct-message/create API 发送消息。

        Args:
            message_: 消息内容，可以是 str, Mapping, Iterable[Mapping],
                'KookMessageSegment', 'KookMessage'。
                将使用 `KookMessage` 进行封装。
            message_type: 消息类型。应该是 GROUP 或者 PERSON。
            id_: 发送对象的 ID ，Kook 用户码或者Kook频道码。

        Returns:
            API 响应。

        Raises:
            TypeError: message_type 不是 'PERSON' 或 'GROUP'。
            ...: 同 `call_api()` 方法。
        """
        if message_type == "PERSON":
            return await self.call_api(
                api="direct-message/create", target_id=id_, content=message_
            )
        elif message_type == "GROUP":
            return await self.call_api(
                api="message/create", target_id=id_, content=message_
            )
        else:
            raise TypeError('message_type must be "PERSON" or "GROUP"')


def _handle_api_result(response: Any) -> Any:
    """
    :说明:

      处理 API 请求返回值。

    :参数:

      * ``response: Response``: API 响应体

    :返回:

        - ``T``: API 调用返回数据

    :异常:

        - ``ActionFailed``: API 调用失败
    """
    result = json.loads(response.content)
    if isinstance(result, dict):
        if result.get("code") != 0:
            raise ActionFailed(response)
        else:
            return result.get("data")



from iamai import Plugin
from iamai.adapter.kook.message import KookMessageSegment


class HalloKook(Plugin):
    async def handle(self) -> None:
        await self.event.reply(KookMessageSegment.Card(self.event.user_id))

    async def rule(self) -> bool:
        if self.event.adapter.name != "kook":
            return False
        # if int(self.event.type) != 9:
        #     return False
        return self.event.message.get_plain_text() == "hello"



"""Bililive 适配器事件。"""

import inspect
from enum import IntEnum
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Type,
    Union,
    Literal,
    TypeVar,
    Optional,
)


from iamai.event import Event

from .message import BililiveMessage

if TYPE_CHECKING:
    from .message import T_BililiveMSG

T_BililiveEvent = TypeVar("T_BililiveEvent", bound="BililiveEvent")


# go-common\app\service\main\broadcast\model\operation.go
class Operation(IntEnum):
    HANDSHAKE = 0
    HANDSHAKE_REPLY = 1
    HEARTBEAT = 2
    HEARTBEAT_REPLY = 3
    SEND_MSG = 4
    SEND_MSG_REPLY = 5
    DISCONNECT_REPLY = 6
    AUTH = 7
    AUTH_REPLY = 8
    RAW = 9
    PROTO_READY = 10
    PROTO_FINISH = 11
    CHANGE_ROOM = 12
    CHANGE_ROOM_REPLY = 13
    REGISTER = 14
    REGISTER_REPLY = 15
    UNREGISTER = 16
    UNREGISTER_REPLY = 17
    # B站业务自定义OP
    # MinBusinessOp = 1000
    # MaxBusinessOp = 10000


class DanmakuPosition(IntEnum):
    TOP = (5,)
    BOTTOM = (4,)
    NORMAL = 1


class BililiveEvent(Event["BililiveAdapter"]):
    """Blilive 适配器事件类。"""

    __event__ = ""
    cmd: str


class MessageEvent(BililiveEvent):
    """消息事件"""

    __event__ = "message"
    post_type: Literal["message"] = "message"
    sub_type: str
    message: BililiveMessage
    session_id: str

    def __repr__(self) -> str:
        return f'Event<{self.type}>: "{self.message}"'

    def get_plain_text(self) -> str:
        return self.message.get_plain_text()

    async def reply(self, msg: "T_BililiveMSG") -> Dict[str, Any]:
        raise NotImplementedError


class Danmu_msg(MessageEvent):
    """弹幕"""

    __event__ = "message.danmu_msg"
    message_type: Literal["danmu_msg"]
    info: List[Any]

    async def reply(self, msg: "T_BililiveMSG") -> Dict[str, Any]:
        return await self.adapter.send(danmaku=msg)


class Super_chat_message(MessageEvent):
    """醒目留言"""

    __event__ = "message.super_chat_message"
    message_type: Literal["super_chat_message"]
    data: Dict[str, Any]
    duration: int


class NoticeEvent(Event):
    __event__ = "notice"


class Combo_send(NoticeEvent):
    """连击礼物"""

    __event__ = "notice.combo_send"
    data: Dict[Any, Any]
    notice_type: Literal["combo_send"]


class Send_gift(NoticeEvent):
    """投喂礼物"""

    __event__ = "notice.send_gift"
    data: Dict[Any, Any]
    notice_type: Literal["send_gift"]


class Common_notice_danmaku(NoticeEvent):
    """限时任务(系统通知的)"""

    __event__ = "notice.common_notice_danmaku"
    data: Dict[Any, Any]
    notice_type: Literal["common_notice_danmaku"]


class Entry_effect(NoticeEvent):
    """舰长进房"""

    __event__ = "notice.entry_effect"
    data: Dict[Any, Any]
    notice_type: Literal["entry_effect"]


class Interact_word(NoticeEvent):
    """普通进房消息"""

    __event__ = "notice_interact_word"
    data: Dict[Any, Any]
    notice_type: Literal["notice_interact_word"]


class Guard_buy(NoticeEvent):
    """上舰"""

    __event__ = "notice.guard_buy"
    data: Dict[Any, Any]
    notice_type: Literal["guard_buy"]


class User_toast_msg(NoticeEvent):
    """续费舰长"""

    __event__ = "notice.user_toast_msg"
    data: Dict[Any, Any]
    notice_type: Literal["user_toast_msg"]


class Notice_msg(NoticeEvent):
    """在本房间续费了舰长"""

    __event__ = "notice.notice_msg"
    id: int
    name: str
    full: Dict[str, Any]
    half: Dict[str, Any]
    side: Dict[str, Any]
    scatter: Dict[str, int]
    roomid: int
    real_roomid: int
    msg_common: int
    msg_self: str
    link_url: str
    msg_type: int
    shield_uid: int
    business_id: str
    marquee_id: str
    notice_type: Union[Literal["notice_msg"], int]


class Like_info_v3_click(NoticeEvent):
    """点赞"""

    __event__ = "notice.like_info_v3_click"
    data: Dict[Any, Any]
    notice_type: Literal["like_info_v3_click"]


class Like_info_v3_update(NoticeEvent):
    """总点赞数"""

    __event__ = "notice.like_info_v3_update"
    data: Dict[Any, Any]
    notice_type: Literal["like_info_v3_update"]


class Online_rank_count(NoticeEvent):
    """在线等级统计"""

    __event__ = "notice.online_rank_count"
    data: Dict[Any, Any]
    notice_type: Literal["online_rank_count"]


class Online_rank_v2(NoticeEvent):
    """在线等级榜"""

    __event__ = "notice.online_rank_v2"
    data: Dict[Any, Any]
    notice_type: Literal["online_rank_v2"]


class Popular_rank_changed(NoticeEvent):
    __event__ = "notice.popular_rank_changed"
    data: Dict[Any, Any]
    notice_type: Literal["popular_rank_changed"]


class Room_change(NoticeEvent):
    """房间信息变动(分区、标题等)"""

    __event__ = "notice.room_change"
    data: Dict[Any, Any]
    notice_type: Literal["room_change"]


class Room_real_time_message_update(NoticeEvent):
    """房间数据"""

    __event__ = "notice.room_real_time_message_update"
    data: Dict[Any, Any]
    notice_type: Literal["room_real_time_message_update"]


class Watched_change(NoticeEvent):
    """直播间观看人数"""

    __event__ = "notice.watched_change"
    data: Dict[Any, Any]
    notice_type: Literal["watched_change"]


class Stop_live_room_list(NoticeEvent):
    """下播列表"""

    __event__ = "notice.stop_live_room_list"
    data: Dict[Any, Any]
    room_id_list: List[int]
    notice_type: Literal["stop_live_room_list"]


class Anchor_lot_start(NoticeEvent):
    """天选之人开始"""

    __event__ = "notice.anchor_lot_start"
    data: Dict[Any, Any]
    notice_type: Literal["anchor_lot_start"]

    def get_anchor_lot_info(self):
        """获取天选之人的相关信息"""
        return {
            "award_name": self.data["award_name"],
            "danmu": self.data["danmu"],
            "gift_name": self.data["gift_name"],
        }


class Anchor_lot_award(NoticeEvent):
    """天选之人结果"""

    __event__ = "notice.anchor_lot_award"
    data: Dict[Any, Any]
    notice_type: Literal["anchor_lot_award"]

    def winner_info(self):
        """获取中奖人信息"""
        return self.data["award_users"]


# 事件类映射
_bililive_events = {
    model.__event__: model
    for model in globals().values()
    if inspect.isclass(model) and issubclass(model, BililiveEvent)
}


def get_event_class(
    post_type: str, event_type: str, sub_type: Optional[str] = None
) -> Type[T_BililiveEvent]:  # type: ignore
    if sub_type is None:
        return _bililive_events[".".join((post_type, event_type))]  # type: ignore
    return (
        _bililive_events.get(".".join((post_type, event_type, sub_type)))
        or _bililive_events[".".join((post_type, event_type))]
    )  # type: ignore



EVENT_DESCRIPTIONS = {
    "commit_comment": '{comment[user][login]} commented on "{comment[commit_id]}" in "{repository[full_name]}"\n{comment[body]}',
    "create": "{sender[login]} created {ref_type} ({ref}) in "
    "{repository[full_name]}",
    "delete": "{sender[login]} deleted {ref_type} ({ref}) in "
    "{repository[full_name]}",
    "deployment": "{sender[login]} deployed {deployment[ref]} to "
    "{deployment[environment]} in {repository[full_name]}",
    "deployment_status": "deployment of {deployement[ref]} to "
    "{deployment[environment]} "
    "{deployment_status[state]} in "
    "{repository[full_name]}",
    "fork": '"{forkee[owner][login]}" forked "{forkee[name]}" (Total {repository[forks_count]} forkee)',
    "gollum": "{sender[login]} edited wiki pages in {repository[full_name]}",
    "issue_comment": {
        "created": '{sender[login]} commented on issue #{issue[number]} in "{repository[full_name]}"\n{comment[body]}'
    },
    "issues": {
        "opened": '{sender[login]} {action} issue #{issue[number]} in "{repository[full_name]}"\n"{issue[title]}"\n{issue[body]}',
        "closed": '{sender[login]} {action} issue #{issue[number]} in "{repository[full_name]}"',
    },
    "member": "{sender[login]} {action} member {member[login]} in "
    "{repository[full_name]}",
    "membership": "{sender[login]} {action} member {member[login]} to team "
    "{team[name]} in {repository[full_name]}",
    "page_build": "{sender[login]} built pages in {repository[full_name]}",
    "ping": "ping from {sender[login]}",
    "public": "{sender[login]} publicized {repository[full_name]}",
    "pull_request": "{sender[login]} {action} pull #{pull_request[number]} in "
    "{repository[full_name]}",
    "pull_request_review": "{sender[login]} {action} {review[state]} "
    "review on pull #{pull_request[number]} in "
    "{repository[full_name]}",
    "pull_request_review_comment": "{comment[user][login]} {action} comment "
    "on pull #{pull_request[number]} in "
    "{repository[full_name]}",
    "push": '{pusher[name]} pushed {ref} in "{repository[full_name]}"\n{commits[0][message]}',
    "release": '{release[author][login]} {action} {release[tag_name]} in "{repository[full_name]}"',
    "repository": "{sender[login]} {action} repository " "{repository[full_name]}",
    "status": "{sender[login]} set {sha} status to {state} in "
    "{repository[full_name]}",
    "team_add": "{sender[login]} added repository {repository[full_name]} to "
    "team {team[name]}",
    "watch": '{sender[login]} {action} watch in repository "{repository[full_name]}"(Total {repository[stargazers_count]} stargazers)',
}



"""Kook 适配器配置。"""

from typing import Literal

from iamai.config import ConfigModel


class Config(ConfigModel):
    """Kook 配置类，将在适配器被加载时被混入到机器人主配置中。

    Attributes:
        adapter_type: 适配器类型，需要和协议端配置相同。
        reconnect_interval: 重连等待时间。
        api_timeout: 进行 API 调用时等待返回响应的超时时间。
        access_token: 鉴权密钥。
        compress: 是否启用压缩，默认为 0，(建议)不启用。
        show_raw: 是否显示原始数据，默认为 False，不显示。
    """

    __config_name__ = "kook"
    adapter_type: Literal["ws"] = "ws"
    reconnect_interval: int = 3
    api_timeout: int = 1000
    access_token: str = ""
    compress: Literal[0, 1] = 0
    show_raw: bool = False
    report_self_message: bool = False



"""DingTalk 适配器事件。"""

import time
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Union
from typing_extensions import Self

from pydantic import BaseModel, Field

from iamai.event import MessageEvent

from .exceptions import WebhookExpiredError
from .message import DingTalkMessage

if TYPE_CHECKING:
    pass

__all__ = ["UserInfo", "Text", "DingTalkEvent"]


class UserInfo(BaseModel):
    """用户信息"""

    dingtalkId: str
    staffId: Optional[str] = None


class Text(BaseModel):
    """文本消息"""

    content: str


class DingTalkEvent(MessageEvent["DingTalkAdapter"]):
    """DingTalk 事件基类"""

    type: Optional[str] = Field(alias="msgtype")

    msgtype: str
    msgId: str
    createAt: str
    conversationType: Literal["1", "2"]
    conversationId: str
    conversationTitle: Optional[str] = None
    senderId: str
    senderNick: str
    senderCorpId: Optional[str] = None
    sessionWebhook: str
    sessionWebhookExpiredTime: int
    isAdmin: Optional[bool] = None
    chatbotCorpId: Optional[str] = None
    isInAtList: Optional[bool] = None
    senderStaffId: Optional[str] = None
    chatbotUserId: str
    atUsers: List[UserInfo]
    text: Text

    response_msg: Union[None, str, Dict[str, Any], DingTalkMessage] = None
    response_at: Union[None, Dict[str, Any], DingTalkMessage] = None

    @property
    def message(self) -> DingTalkMessage:
        """返回 message 字段。"""
        return DingTalkMessage.text(self.text.content)

    def get_plain_text(self) -> str:
        """获取消息的纯文本内容。

        Returns:
            消息的纯文本内容。
        """
        return self.message.get_plain_text()

    async def reply(
        self,
        message: Union[str, Dict[str, Any], DingTalkMessage],
        at: Union[None, Dict[str, Any], DingTalkMessage] = None,
    ) -> Dict[str, Any]:
        """回复消息。

        Args:
            message: 回复消息的内容，可以是 `str`, `Dict` 或 `DingTalkMessage`。
            at: 回复消息时 At 的对象，必须时 at 类型的 `DingTalkMessage`，或者符合标准的 `Dict`。

        Returns:
            调用 Webhook 地址后钉钉服务器的响应。

        Raises:
            WebhookExpiredError: 当前事件的 Webhook 地址已经过期。
            ...: 同 `DingTalkAdapter.send()` 方法。
        """
        if self.sessionWebhookExpiredTime > time.time() * 1000:
            return await self.adapter.send(
                webhook=self.sessionWebhook,
                conversation_type=self.conversationType,
                msg=message,
                at=at,
            )
        raise WebhookExpiredError

    async def is_same_sender(self, other: Self) -> bool:
        """判断自身和另一个事件是否是同一个发送者。

        Args:
            other: 另一个事件。

        Returns:
            是否是同一个发送者。
        """
        return self.senderId == other.senderId



"""Bot Module

The basic module of iamai, each iamai robot is a ``Bot()`` instance.
"""

import asyncio
import json
import pkgutil
import signal
import sys
import threading
import time
from collections import defaultdict
from contextlib import AsyncExitStack
from itertools import chain
from pathlib import Path
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    overload,
)

from pydantic import ValidationError, create_model

from .adapter import Adapter
from .config import AdapterConfig, ConfigModel, MainConfig, PluginConfig
from .dependencies import solve_dependencies
from .event import Event
from .exceptions import (
    GetEventTimeout,
    LoadModuleError,
    SkipException,
    StopException,
)
from .log import logger
from .plugin import Plugin, PluginLoadType
from .typing import AdapterHook, AdapterT, BotHook, EventHook, EventT
from .utils import (
    ModulePathFinder,
    get_classes_from_module_name,
    is_config_class,
    samefile,
    wrap_get_func,
)

if sys.version_info >= (3, 11):  # pragma: no cover
    import tomllib
else:  # pragma: no cover
    import tomli as tomllib


__all__ = ["Bot"]

HANDLED_SIGNALS = (
    signal.SIGINT,  # Unix signal 2. Sent by Ctrl+C.
    signal.SIGTERM,  # Unix signal 15. Sent by `kill <pid>`.
)


class Bot:
    """iamai ``Bot`` class object, defines the basic functionality  for interacting.

    Read and save configuration ``Config``, load adapters ``Adapter`` and plugins ``Plugin``, then distribute the events.

    Attributes:
        config: Bot configuration.
        should_exit: Whether the bot should enter the ready-to-exit state.
        adapters: List of currently loaded adapters.
        plugins_priority_dict: Plugin priority dictionary.
        plugin_state: Plugin status.
        global_state: Global status.
    """

    config: MainConfig
    should_exit: asyncio.Event  # pyright: ignore[reportUninitializedInstanceVariable]
    adapters: List[Adapter[Any, Any]]
    plugins_priority_dict: Dict[int, List[Type[Plugin[Any, Any, Any]]]]
    plugin_state: Dict[str, Any]
    global_state: Dict[Any, Any]

    _condition: asyncio.Condition  # Condition used to handle get # pyright: ignore[reportUninitializedInstanceVariable]
    _current_event: Optional[Event[Any]]  # Event currently pending

    _restart_flag: bool  # Restart flag
    _module_path_finder: ModulePathFinder  # Module metapath finder for finding plugins
    _raw_config_dict: Dict[str, Any]  # Original configuration dictionary
    _adapter_tasks: Set[
        "asyncio.Task[None]"
    ]  # Adapter task collection, used to hold references to adapter tasks
    _handle_event_tasks: Set[
        "asyncio.Task[None]"
    ]  # Event handling task, used to keep a reference to the adapter task

    # The following properties are not cleared on reboot
    _config_file: Optional[str]  # Configuration file
    _config_dict: Optional[Dict[str, Any]]  # Configuration dictionary
    _hot_reload: bool  # Hot-Reload

    _extend_plugins: List[
        Union[Type[Plugin[Any, Any, Any]], str, Path]
    ]  # A list of plugins loaded programmatically using the ``load_plugins()`` method
    _extend_plugin_dirs: List[
        Path
    ]  # List of plugin paths loaded programmatically using the ``load_plugins_from_dirs()`` method
    _extend_adapters: List[
        Union[Type[Adapter[Any, Any]], str]
    ]  # A list of adapters loaded programmatically using the ``load_adapter()`` method
    _bot_run_hooks: List[BotHook]
    _bot_exit_hooks: List[BotHook]
    _adapter_startup_hooks: List[AdapterHook]
    _adapter_run_hooks: List[AdapterHook]
    _adapter_shutdown_hooks: List[AdapterHook]
    _event_preprocessor_hooks: List[EventHook]
    _event_postprocessor_hooks: List[EventHook]

    def __init__(
        self,
        *,
        config_file: Optional[str] = "config.toml",
        config_dict: Optional[Dict[str, Any]] = None,
        hot_reload: bool = False,
    ) -> None:
        """Initialize iamai, read configuration files, create configurations, load adapters and plug-ins.

        Args:
            config_file: Configuration file, if not specified, the default ``config.toml`` will be used.
                If specified as ``None``, the configuration file will not be loaded.
            config_dict: Configuration dictionary, default is ``None``.
                If a dictionary is specified, the ``config_file`` configuration will be ignored and the configuration file will no longer be read.
            hot_reload: hot reload.
                When enabled, plugin files in ``plugin_dir`` will be automatically checked for updates and automatically reloaded when updated.
        """
        self.config = MainConfig()
        self.plugins_priority_dict = defaultdict(list)
        self.plugin_state = defaultdict(lambda: None)
        self.global_state = {}

        self.adapters = []
        self._current_event = None
        self._restart_flag = False
        self._module_path_finder = ModulePathFinder()
        self._raw_config_dict = {}
        self._adapter_tasks = set()
        self._handle_event_tasks = set()

        self._config_file = config_file
        self._config_dict = config_dict
        self._hot_reload = hot_reload

        self._extend_plugins = []
        self._extend_plugin_dirs = []
        self._extend_adapters = []
        self._bot_run_hooks = []
        self._bot_exit_hooks = []
        self._adapter_startup_hooks = []
        self._adapter_run_hooks = []
        self._adapter_shutdown_hooks = []
        self._event_preprocessor_hooks = []
        self._event_postprocessor_hooks = []

        sys.meta_path.insert(0, self._module_path_finder)

    @property
    def plugins(self) -> List[Type[Plugin[Any, Any, Any]]]:
        """List of currently loaded plugins."""
        return list(chain(*self.plugins_priority_dict.values()))

    def run(self) -> None:
        """Run iamai, monitor and intercept system exit signals, and update robot configuration."""
        self._restart_flag = True
        while self._restart_flag:
            self._restart_flag = False
            asyncio.run(self._run())
            if self._restart_flag:
                self._load_plugins_from_dirs(*self._extend_plugin_dirs)
                self._load_plugins(*self._extend_plugins)
                self._load_adapters(*self._extend_adapters)

    def restart(self) -> None:
        """Exit and rerun iamai."""
        logger.info("Restarting iamai...")
        self._restart_flag = True
        self.should_exit.set()

    async def _run(self) -> None:
        """Run iamai."""
        self.should_exit = asyncio.Event()
        self._condition = asyncio.Condition()

        # Monitor and intercept system exit signals to complete some aftermath work before closing the program
        if threading.current_thread() is threading.main_thread():  # pragma: no cover
            # Signals can only be processed in the main thread
            try:
                loop = asyncio.get_running_loop()
                for sig in HANDLED_SIGNALS:
                    loop.add_signal_handler(sig, self._handle_exit)
            except NotImplementedError:
                # add_signal_handler is only available under Unix, below for Windows
                for sig in HANDLED_SIGNALS:
                    signal.signal(sig, self._handle_exit)

        # Load configuration file
        self._reload_config_dict()

        # Load plugins and adapters
        self._load_plugins_from_dirs(*self.config.bot.plugin_dirs)
        self._load_plugins(*self.config.bot.plugins)
        self._load_adapters(*self.config.bot.adapters)
        self._update_config()

        # Run iamai
        logger.info("Running iamai...")

        hot_reload_task = None
        if self._hot_reload:  # pragma: no cover
            hot_reload_task = asyncio.create_task(self._run_hot_reload())

        for bot_run_hook_func in self._bot_run_hooks:
            await bot_run_hook_func(self)

        try:
            for _adapter in self.adapters:
                for adapter_startup_hook_func in self._adapter_startup_hooks:
                    await adapter_startup_hook_func(_adapter)
                try:
                    await _adapter.startup()
                except Exception as e:
                    self.error_or_exception(f"Startup adapter {_adapter!r} failed:", e)

            for _adapter in self.adapters:
                for adapter_run_hook_func in self._adapter_run_hooks:
                    await adapter_run_hook_func(_adapter)
                _adapter_task = asyncio.create_task(_adapter.safe_run())
                self._adapter_tasks.add(_adapter_task)
                _adapter_task.add_done_callback(self._adapter_tasks.discard)

            await self.should_exit.wait()

            if hot_reload_task is not None:  # pragma: no cover
                await hot_reload_task
        finally:
            for _adapter in self.adapters:
                for adapter_shutdown_hook_func in self._adapter_shutdown_hooks:
                    await adapter_shutdown_hook_func(_adapter)
                await _adapter.shutdown()

            while self._adapter_tasks:
                await asyncio.sleep(0)

            for bot_exit_hook_func in self._bot_exit_hooks:
                await bot_exit_hook_func(self)

            self.adapters.clear()
            self.plugins_priority_dict.clear()
            self._module_path_finder.path.clear()

    def _remove_plugin_by_path(
        self, file: Path
    ) -> List[Type[Plugin[Any, Any, Any]]]:  # pragma: no cover
        """Remove loaded plugins based on path."""
        removed_plugins: List[Type[Plugin[Any, Any, Any]]] = []
        for plugins in self.plugins_priority_dict.values():
            _removed_plugins = list(
                filter(
                    lambda x: x.__plugin_load_type__ != PluginLoadType.CLASS
                    and x.__plugin_file_path__ is not None
                    and samefile(x.__plugin_file_path__, file),
                    plugins,
                )
            )
            removed_plugins.extend(_removed_plugins)
            for plugin_ in _removed_plugins:
                plugins.remove(plugin_)
                logger.info(
                    "Succeeded to remove plugin "
                    f'"{plugin_.__name__}" from file "{file}"'
                )
        return removed_plugins

    async def _run_hot_reload(self) -> None:  # pragma: no cover
        """Hot reload."""
        try:
            from watchfiles import Change, awatch
        except ImportError:
            logger.warning(
                'Hot reload needs to install "watchfiles", try "pip install watchfiles"'
            )
            return

        logger.info("Hot reload is working!")
        async for changes in awatch(
            *(
                x.resolve()
                for x in set(self._extend_plugin_dirs)
                .union(self.config.bot.plugin_dirs)
                .union(
                    {Path(self._config_file)}
                    if self._config_dict is None and self._config_file is not None
                    else set()
                )
            ),
            stop_event=self.should_exit,
        ):
            # Processed in the order of Change.deleted, Change.modified, Change.added
            # To ensure that when renaming occurs, deletions are processed first and then additions are processed
            for change_type, file_ in sorted(changes, key=lambda x: x[0], reverse=True):
                file = Path(file_)
                # Change configuration file
                if (
                    self._config_file is not None
                    and samefile(self._config_file, file)
                    and change_type == change_type.modified
                ):
                    logger.info(f'Reload config file "{self._config_file}"')
                    old_config = self.config
                    self._reload_config_dict()
                    if (
                        self.config.bot != old_config.bot
                        or self.config.adapter != old_config.adapter
                    ):
                        self.restart()
                    continue

                # Change plugin folder
                if change_type == Change.deleted:
                    # Special handling for deletion operations
                    if file.suffix != ".py":
                        file = file / "__init__.py"
                else:
                    if file.is_dir() and (file / "__init__.py").is_file():
                        # When a new directory is added and this directory contains the ``__init__.py`` file
                        # It means that what happens at this time is that a Python package is added, and the ``__init__.py`` file of this package is deemed to be added
                        file = file / "__init__.py"
                    if not (file.is_file() and file.suffix == ".py"):
                        continue

                if change_type == Change.added:
                    logger.info(f"Hot reload: Added file: {file}")
                    self._load_plugins(
                        Path(file), plugin_load_type=PluginLoadType.DIR, reload=True
                    )
                    self._update_config()
                    continue
                if change_type == Change.deleted:
                    logger.info(f"Hot reload: Deleted file: {file}")
                    self._remove_plugin_by_path(file)
                    self._update_config()
                elif change_type == Change.modified:
                    logger.info(f"Hot reload: Modified file: {file}")
                    self._remove_plugin_by_path(file)
                    self._load_plugins(
                        Path(file), plugin_load_type=PluginLoadType.DIR, reload=True
                    )
                    self._update_config()

    def _update_config(self) -> None:
        """Updated config to incorporate Config from Plugin and Adapter."""

        def update_config(
            source: Union[List[Type[Plugin[Any, Any, Any]]], List[Adapter[Any, Any]]],
            name: str,
            base: Type[ConfigModel],
        ) -> Tuple[Type[ConfigModel], ConfigModel]:
            config_update_dict: Dict[str, Any] = {}
            for i in source:
                config_class = getattr(i, "Config", None)
                if is_config_class(config_class):
                    default_value: Any
                    try:
                        default_value = config_class()
                    except ValidationError:
                        default_value = ...
                    config_update_dict[config_class.__config_name__] = (
                        config_class,
                        default_value,
                    )
            config_model = create_model(name, **config_update_dict, __base__=base)
            return config_model, config_model()

        self.config = create_model(
            "Config",
            plugin=update_config(self.plugins, "PluginConfig", PluginConfig),
            adapter=update_config(self.adapters, "AdapterConfig", AdapterConfig),
            __base__=MainConfig,
        )(**self._raw_config_dict)
        # Update the level of logging
        logger.remove()
        logger.add(sys.stderr, level=self.config.bot.log.level)

    def _reload_config_dict(self) -> None:
        """Reload the configuration file."""
        self._raw_config_dict = {}

        if self._config_dict is not None:
            self._raw_config_dict = self._config_dict
        elif self._config_file is not None:
            try:
                with Path(self._config_file).open("rb") as f:
                    if self._config_file.endswith(".json"):
                        self._raw_config_dict = json.load(f)
                    elif self._config_file.endswith(".toml"):
                        self._raw_config_dict = tomllib.load(f)
                    else:
                        self.error_or_exception(
                            "Read config file failed:",
                            OSError("Unable to determine config file type"),
                        )
            except OSError as e:
                self.error_or_exception("Can not open config file:", e)
            except (ValueError, json.JSONDecodeError, tomllib.TOMLDecodeError) as e:
                self.error_or_exception("Read config file failed:", e)

        try:
            self.config = MainConfig(**self._raw_config_dict)
        except ValidationError as e:
            self.config = MainConfig()
            self.error_or_exception("Config dict parse error:", e)
        self._update_config()

    def reload_plugins(self) -> None:
        """Manually reload all plugins."""
        self.plugins_priority_dict.clear()
        self._load_plugins(*self.config.bot.plugins)
        self._load_plugins_from_dirs(*self.config.bot.plugin_dirs)
        self._load_plugins(*self._extend_plugins)
        self._load_plugins_from_dirs(*self._extend_plugin_dirs)
        self._update_config()

    def _handle_exit(self, *_args: Any) -> None:  # pragma: no cover
        """When the robot receives the exit signal, it will handle it according to the situation."""
        logger.info("Stopping iamai...")
        if self.should_exit.is_set():
            logger.warning("Force Exit iamai...")
            sys.exit()
        else:
            self.should_exit.set()

    async def handle_event(
        self,
        current_event: Event[Any],
        *,
        handle_get: bool = True,
        show_log: bool = True,
    ) -> None:
        """Called by the adapter object, distributes events to all plug-ins according to priority, and handles plug-in signals such as ``stop`` and ``skip``.

        This method should not be called manually by the user.

        Args:
            current_event: The currently pending ``Event``.
            handle_get: Whether the current event can be captured by the get method, the default is ``True``.
            show_log: Whether to display in the log, the default is ``True``.
        """
        if show_log:
            logger.info(
                f"Adapter {current_event.adapter.name} received: {current_event!r}"
            )

        if handle_get:
            _handle_event_task = asyncio.create_task(self._handle_event())
            self._handle_event_tasks.add(_handle_event_task)
            _handle_event_task.add_done_callback(self._handle_event_tasks.discard)
            await asyncio.sleep(0)
            async with self._condition:
                self._current_event = current_event
                self._condition.notify_all()
        else:
            _handle_event_task = asyncio.create_task(self._handle_event(current_event))
            self._handle_event_tasks.add(_handle_event_task)
            _handle_event_task.add_done_callback(self._handle_event_tasks.discard)

    async def _handle_event(self, current_event: Optional[Event[Any]] = None) -> None:
        if current_event is None:
            async with self._condition:
                await self._condition.wait()
                assert self._current_event is not None
                current_event = self._current_event
            if current_event.__handled__:
                return

        for _hook_func in self._event_preprocessor_hooks:
            await _hook_func(current_event)

        for plugin_priority in sorted(self.plugins_priority_dict.keys()):
            logger.debug(
                f"Checking for matching plugins with priority {plugin_priority!r}"
            )
            stop = False
            for plugin in self.plugins_priority_dict[plugin_priority]:
                try:
                    async with AsyncExitStack() as stack:
                        _plugin = await solve_dependencies(
                            plugin,
                            use_cache=True,
                            stack=stack,
                            dependency_cache={
                                Bot: self,
                                Event: current_event,
                            },
                        )
                        if _plugin.name not in self.plugin_state:
                            plugin_state = _plugin.__init_state__()
                            if plugin_state is not None:
                                self.plugin_state[_plugin.name] = plugin_state
                        if await _plugin.rule():
                            logger.info(f"Event will be handled by {_plugin!r}")
                            try:
                                await _plugin.handle()
                            finally:
                                if _plugin.block:
                                    stop = True
                except SkipException:
                    # The plug-in requires that it skips itself and continues the current event propagation
                    continue
                except StopException:
                    # Plugin requires stopping current event propagation
                    stop = True
                except Exception as e:
                    self.error_or_exception(f'Exception in plugin "{plugin}":', e)
            if stop:
                break

        for _hook_func in self._event_postprocessor_hooks:
            await _hook_func(current_event)

        logger.info("Event Finished")

    @overload
    async def get(
        self,
        func: Optional[Callable[[Event[Any]], Union[bool, Awaitable[bool]]]] = None,
        *,
        event_type: None = None,
        adapter_type: None = None,
        max_try_times: Optional[int] = None,
        timeout: Optional[Union[int, float]] = None,
    ) -> Event[Any]: ...

    @overload
    async def get(
        self,
        func: Optional[Callable[[EventT], Union[bool, Awaitable[bool]]]] = None,
        *,
        event_type: None = None,
        adapter_type: Type[Adapter[EventT, Any]],
        max_try_times: Optional[int] = None,
        timeout: Optional[Union[int, float]] = None,
    ) -> EventT: ...

    @overload
    async def get(
        self,
        func: Optional[Callable[[EventT], Union[bool, Awaitable[bool]]]] = None,
        *,
        event_type: Type[EventT],
        adapter_type: Optional[Type[AdapterT]] = None,
        max_try_times: Optional[int] = None,
        timeout: Optional[Union[int, float]] = None,
    ) -> EventT: ...

    async def get(
        self,
        func: Optional[Callable[[Any], Union[bool, Awaitable[bool]]]] = None,
        *,
        event_type: Optional[Type[Event[Any]]] = None,
        adapter_type: Optional[Type[Adapter[Any, Any]]] = None,
        max_try_times: Optional[int] = None,
        timeout: Optional[Union[int, float]] = None,
    ) -> Event[Any]:
        """Get events that meet the specified conditions. The coroutine will wait until the adapter receives events that meet the conditions, exceeds the maximum number of events, or times out.

        Args:
            func: Coroutine or function, the function will be automatically packaged as a coroutine for execution.
                Requires an event to be accepted as a parameter and returns a Boolean value. Returns the current event when the coroutine returns ``True``.
                When ``None`` is equivalent to the input coroutine returning true for any event, that is, returning the next event received by the adapter.
            event_type: When specified, only events of the specified type are accepted, taking effect before the func condition. Defaults to ``None``.
            adapter_type: When specified, only events generated by the specified adapter will be accepted, taking effect before the func condition. Defaults to ``None``.
            max_try_times: Maximum number of events.
            timeout: timeout period.

        Returns:
            Returns events that satisfy the condition of ``func``.

        Raises:
            GetEventTimeout: Maximum number of events exceeded or timeout.
        """
        _func = wrap_get_func(func)

        try_times = 0
        start_time = time.time()
        while not self.should_exit.is_set():
            if max_try_times is not None and try_times > max_try_times:
                break
            if timeout is not None and time.time() - start_time > timeout:
                break

            async with self._condition:
                if timeout is None:
                    await self._condition.wait()
                else:
                    try:
                        await asyncio.wait_for(
                            self._condition.wait(),
                            timeout=start_time + timeout - time.time(),
                        )
                    except asyncio.TimeoutError:
                        break

                if (
                    self._current_event is not None
                    and not self._current_event.__handled__
                    and (
                        event_type is None
                        or isinstance(self._current_event, event_type)
                    )
                    and (
                        adapter_type is None
                        or isinstance(self._current_event.adapter, adapter_type)
                    )
                    and await _func(self._current_event)
                ):
                    self._current_event.__handled__ = True
                    return self._current_event

                try_times += 1

        raise GetEventTimeout

    def _load_plugin_class(
        self,
        plugin_class: Type[Plugin[Any, Any, Any]],
        plugin_load_type: PluginLoadType,
        plugin_file_path: Optional[str],
    ) -> None:
        """Load a plugin class"""
        priority = getattr(plugin_class, "priority", None)
        if isinstance(priority, int) and priority >= 0:
            for _plugin in self.plugins:
                if _plugin.__name__ == plugin_class.__name__:
                    logger.warning(
                        f'Already have a same name plugin "{_plugin.__name__}"'
                    )
            plugin_class.__plugin_load_type__ = plugin_load_type
            plugin_class.__plugin_file_path__ = plugin_file_path
            self.plugins_priority_dict[priority].append(plugin_class)
            logger.info(
                f'Succeeded to load plugin "{plugin_class.__name__}" '
                f'from class "{plugin_class!r}"'
            )
        else:
            self.error_or_exception(
                f'Load plugin from class "{plugin_class!r}" failed:',
                LoadModuleError(
                    f'Plugin priority incorrect in the class "{plugin_class!r}"'
                ),
            )

    def _load_plugins_from_module_name(
        self,
        module_name: str,
        *,
        plugin_load_type: PluginLoadType,
        reload: bool = False,
    ) -> None:
        """Load plugins from the given module."""
        try:
            plugin_classes = get_classes_from_module_name(
                module_name, Plugin, reload=reload
            )
        except ImportError as e:
            self.error_or_exception(f'Import module "{module_name}" failed:', e)
        else:
            for plugin_class, module in plugin_classes:
                self._load_plugin_class(
                    plugin_class,  # type: ignore
                    plugin_load_type,
                    module.__file__,
                )

    def _load_plugins(
        self,
        *plugins: Union[Type[Plugin[Any, Any, Any]], str, Path],
        plugin_load_type: Optional[PluginLoadType] = None,
        reload: bool = False,
    ) -> None:
        """Load plugins.

        Args:
            *plugins: plug-in class, plug-in module name or plug-in module file path. Type can be ``Type[Plugin]``, ``str`` or ``pathlib.Path``.
                If it is ``Type[Plugin]``, it will be loaded as a plug-in class.
                If it is of type ``str``, it will be loaded as the plug-in module name, and the format is the same as the Python ``import`` statement.
                    For example: ``path.of.plugin``.
                If it is of type ``pathlib.Path``, it will be loaded as the plug-in module file path.
                    For example: ``pathlib.Path("path/of/plugin")``.
            plugin_load_type: Plug-in loading type, if it is ``None``, it will be automatically determined, otherwise the specified type will be used.
            reload: Whether to reload the module.
        """
        for plugin_ in plugins:
            try:
                if isinstance(plugin_, type) and issubclass(plugin_, Plugin):
                    self._load_plugin_class(
                        plugin_, plugin_load_type or PluginLoadType.CLASS, None
                    )
                elif isinstance(plugin_, str):
                    logger.info(f'Loading plugins from module "{plugin_}"')
                    self._load_plugins_from_module_name(
                        plugin_,
                        plugin_load_type=plugin_load_type or PluginLoadType.NAME,
                        reload=reload,
                    )
                elif isinstance(plugin_, Path):
                    logger.info(f'Loading plugins from path "{plugin_}"')
                    if not plugin_.is_file():
                        raise LoadModuleError(  # noqa: TRY301
                            f'The plugin path "{plugin_}" must be a file'
                        )

                    if plugin_.suffix != ".py":
                        raise LoadModuleError(  # noqa: TRY301
                            f'The path "{plugin_}" must endswith ".py"'
                        )

                    plugin_module_name = None
                    for path in self._module_path_finder.path:
                        try:
                            if plugin_.stem == "__init__":
                                if plugin_.resolve().parent.parent.samefile(Path(path)):
                                    plugin_module_name = plugin_.resolve().parent.name
                                    break
                            elif plugin_.resolve().parent.samefile(Path(path)):
                                plugin_module_name = plugin_.stem
                                break
                        except OSError:
                            continue
                    if plugin_module_name is None:
                        rel_path = plugin_.resolve().relative_to(Path().resolve())
                        if rel_path.stem == "__init__":
                            plugin_module_name = ".".join(rel_path.parts[:-1])
                        else:
                            plugin_module_name = ".".join(
                                rel_path.parts[:-1] + (rel_path.stem,)
                            )

                    self._load_plugins_from_module_name(
                        plugin_module_name,
                        plugin_load_type=plugin_load_type or PluginLoadType.FILE,
                        reload=reload,
                    )
                else:
                    raise TypeError(  # noqa: TRY301
                        f"{plugin_} can not be loaded as plugin"
                    )
            except Exception as e:
                self.error_or_exception(f'Load plugin "{plugin_}" failed:', e)

    def load_plugins(
        self, *plugins: Union[Type[Plugin[Any, Any, Any]], str, Path]
    ) -> None:
        """Load the plugin.

        Args:
            *plugins: ``Plugin`` class, plugin module name or plug-in module file path.
                Type can be ``Type[Plugin]``, ``str`` or ``pathlib.Path``.
                If it is ``Type[Plugin]``, it will be loaded as a plug-in class.
                If it is of type ``str``, it will be loaded as the plug-in module name, and the format is the same as the Python ``import`` statement.
                    For example: ``path.of.plugin``.
                If it is of type ``pathlib.Path``, it will be loaded as the plug-in module file path.
                    For example: ``pathlib.Path("path/of/plugin")``.
        """
        self._extend_plugins.extend(plugins)

        return self._load_plugins(*plugins)

    def _load_plugins_from_dirs(self, *dirs: Path) -> None:
        """Load plug-ins from the directory. Plug-ins in modules starting with ``_`` will not be imported. The path can be a relative path or an absolute path.

        Args:
            *dirs: Module paths that store modules containing plugins.
                For example: ``pathlib.Path("path/of/plugins/")`` .
        """
        dir_list = [str(x.resolve()) for x in dirs]
        logger.info(f'Loading plugins from dirs "{", ".join(map(str, dir_list))}"')
        self._module_path_finder.path.extend(dir_list)
        for module_info in pkgutil.iter_modules(dir_list):
            if not module_info.name.startswith("_"):
                self._load_plugins_from_module_name(
                    module_info.name, plugin_load_type=PluginLoadType.DIR
                )

    def load_plugins_from_dirs(self, *dirs: Path) -> None:
        """Load plug-ins from the directory. Plug-ins in modules starting with ``_`` will not be imported. The path can be a relative path or an absolute path.

        Args:
            *dirs: Module paths that store modules containing plugins.
                For example: ``pathlib.Path("path/of/plugins/")`` .
        """
        self._extend_plugin_dirs.extend(dirs)
        self._load_plugins_from_dirs(*dirs)

    def _load_adapters(self, *adapters: Union[Type[Adapter[Any, Any]], str]) -> None:
        """Load adapter.

        Args:
            *adapters: Adapter class or adapter name, type can be ``Type[Adapter]`` or ``str``.
                If it is of type ``Type[Adapter]``, it will be loaded as an adapter class.
                If it is of type ``str``, it will be loaded as the adapter module name, and the format is the same as the Python ``import`` statement.
                    For example: ``path.of.adapter``.
        """
        for adapter_ in adapters:
            adapter_object: Adapter[Any, Any]
            try:
                if isinstance(adapter_, type) and issubclass(adapter_, Adapter):
                    adapter_object = adapter_(self)
                elif isinstance(adapter_, str):
                    adapter_classes = get_classes_from_module_name(adapter_, Adapter)
                    if not adapter_classes:
                        raise LoadModuleError(  # noqa: TRY301
                            f"Can not find Adapter class in the {adapter_} module"
                        )
                    if len(adapter_classes) > 1:
                        raise LoadModuleError(  # noqa: TRY301
                            f"More then one Adapter class in the {adapter_} module"
                        )
                    adapter_object = adapter_classes[0][0](self)  # type: ignore
                else:
                    raise TypeError(  # noqa: TRY301
                        f"{adapter_} can not be loaded as adapter"
                    )
            except Exception as e:
                self.error_or_exception(f'Load adapter "{adapter_}" failed:', e)
            else:
                self.adapters.append(adapter_object)
                logger.info(
                    f'Succeeded to load adapter "{adapter_object.__class__.__name__}" '
                    f'from "{adapter_}"'
                )

    def load_adapters(self, *adapters: Union[Type[Adapter[Any, Any]], str]) -> None:
        """Load adapter.

        Args:
            *adapters: Adapter class or adapter name, type can be ``Type[Adapter]`` or ``str``.
                If it is of type ``Type[Adapter]``, it will be loaded as an adapter class.
                If it is of type ``str``, it will be loaded as the adapter module name, and the format is the same as the Python ``import`` statement.
                    For example: ``path.of.adapter``.
        """
        self._extend_adapters.extend(adapters)
        self._load_adapters(*adapters)

    @overload
    def get_adapter(self, adapter: str) -> Adapter[Any, Any]: ...

    @overload
    def get_adapter(self, adapter: Type[AdapterT]) -> AdapterT: ...

    def get_adapter(
        self, adapter: Union[str, Type[AdapterT]]
    ) -> Union[Adapter[Any, Any], AdapterT]:
        """Get the loaded adapter by name or adapter class.

        Args:
            adapter: adapter name or adapter class.

        Returns:
            The obtained adapter object.

        Raises:
            LookupError: No adapter object with this name found.
        """
        for _adapter in self.adapters:
            if isinstance(adapter, str):
                if _adapter.name == adapter:
                    return _adapter
            elif isinstance(_adapter, adapter):
                return _adapter
        raise LookupError(f'Can not find adapter named "{adapter}"')

    def get_plugin(self, name: str) -> Type[Plugin[Any, Any, Any]]:
        """Get the loaded plugin class by name.

        Args:
            name: plugin name

        Returns:
            The obtained plug-in class.

        Raises:
            LookupError: The plugin class with this name cannot be found.
        """
        for _plugin in self.plugins:
            if _plugin.__name__ == name:
                return _plugin
        raise LookupError(f'Can not find plugin named "{name}"')

    def error_or_exception(
        self, message: str, exception: Exception
    ) -> None:  # pragma: no cover
        """Output error or exception logs based on the current Bot configuration.

        Args:
            message: message.
            exception: Exception.
        """
        if self.config.bot.log.verbose_exception:
            logger.exception(message)
        else:
            logger.error(f"{message} {exception!r}")

    def bot_run_hook(self, func: BotHook) -> BotHook:
        """Register a function when Bot starts.

        Args:
            func: the registered function.

        Returns:
            The registered function.
        """
        self._bot_run_hooks.append(func)
        return func

    def bot_exit_hook(self, func: BotHook) -> BotHook:
        """Register a function when the Bot exits.

        Args:
            func: the registered function.

        Returns:
            The registered function.
        """
        self._bot_exit_hooks.append(func)
        return func

    def adapter_startup_hook(self, func: AdapterHook) -> AdapterHook:
        """Register a function during adapter initialization.

        Args:
            func: the registered function.

        Returns:
            The registered function.
        """
        self._adapter_startup_hooks.append(func)
        return func

    def adapter_run_hook(self, func: AdapterHook) -> AdapterHook:
        """Register an adapter runtime function.

        Args:
            func: the registered function.

        Returns:
            The registered function.
        """
        self._adapter_run_hooks.append(func)
        return func

    def adapter_shutdown_hook(self, func: AdapterHook) -> AdapterHook:
        """Register a function when the adapter is closed.

        Args:
            func: the registered function.

        Returns:
            The registered function.
        """
        self._adapter_shutdown_hooks.append(func)
        return func

    def event_preprocessor_hook(self, func: EventHook) -> EventHook:
        """Register an event preprocessing function.

        Args:
            func: the registered function.

        Returns:
            The registered function.
        """
        self._event_preprocessor_hooks.append(func)
        return func

    def event_postprocessor_hook(self, func: EventHook) -> EventHook:
        """Register a post-event processing function.

        Args:
            func: the registered function.

        Returns:
            The registered function.
        """
        self._event_postprocessor_hooks.append(func)
        return func



"""Red 适配器事件。"""

import inspect
from enum import IntEnum
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Dict, Type, Union, Literal, TypeVar, Optional

from pydantic import BaseModel

from iamai.event import Event

from .message import T_RedMSG, RedMessage

if TYPE_CHECKING:
    from .message import T_RedMSG

T_RedEvent = TypeVar("T_RedEvent", bound="RedEvent")


class RedEvent(Event["RedAdapter"]):
    """Red 事件基类"""

    __event__ = ""
    # type = Optional[str] = Field(alias="post_type")
    post_type: Literal["message", "notice", "request", "meta_event"]


class EmojiAd(BaseModel):
    url: str
    desc: str


class EmojiMall(BaseModel):
    packageId: int
    emojiId: int


class EmojiZplan(BaseModel):
    actionId: int
    actionName: str
    actionType: int
    playerNumber: int
    peerUid: str
    bytesReserveInfo: str


class ThumbPath(BaseModel): ...


class TextElement(BaseModel):
    content: str
    atType: Literal[0, 1, 2]
    atUid: str
    atTinyId: str
    atNtUid: str
    subElementType: int
    atChannelId: str
    atRoleId: str
    atRoleColor: int
    atRoleName: str
    needNotify: int


class RoleInfo(BaseModel):
    roleId: str
    name: str
    color: int


class XMLElement(BaseModel):
    busiType: str
    busiId: str
    c2cType: int
    serviceType: int
    ctrlFlag: int
    content: str
    templId: str
    seqId: str
    templParam: Any
    pbReserv: str
    members: Any


class PicElement(BaseModel):
    picSubType: int
    fileName: str
    fileSize: str
    picWidth: int
    picHeight: int
    original: bool
    md5HexStr: str
    sourcePath: str
    thumbPath: ThumbPath
    transferStatus: int
    progress: int
    picType: int
    invalidState: int
    fileUuid: str
    fileSubId: str
    thumbFileSize: int
    summary: str
    emojiAd: EmojiAd
    emojiMall: EmojiMall
    emojiZplan: EmojiZplan


class Element(BaseModel):
    elementType: int
    elementId: str
    extBufForUI: str
    picElement: Optional[PicElement]
    textElement: Optional[TextElement]
    arkElement: Optional[Any]
    avRecordElement: Optional[Any]
    calendarElement: Optional[Any]
    faceElement: Optional[Any]
    fileElement: Optional[Any]
    giphyElement: Optional[Any]

    class grayTipElement:
        xmlElement: XMLElement
        aioOpGrayTipElement: Optional[Any]
        blockGrayTipElement: Optional[Any]
        buddyElement: Optional[Any]
        buddyNotifyElement: Optional[Any]
        emojiReplyElement: Optional[Any]
        essenceElement: Optional[Any]
        feedMsgElement: Optional[Any]
        fileReceiptElement: Optional[Any]
        groupElement: Optional[Any]
        groupNotifyElement: Optional[Any]
        jsonGrayTipElement: Optional[Any]
        localGrayTipElement: Optional[Any]
        proclamationElement: Optional[Any]
        revokeElement: Optional[Any]
        subElementType: Optional[Any]

    inlineKeyboardElement: Optional[Any]
    liveGiftElement: Optional[Any]
    markdownElement: Optional[Any]
    marketFaceElement: Optional[Any]
    multiForwardMsgElement: Optional[Any]
    pttElement: Optional[Any]
    replyElement: Optional[Any]
    structLongMsgElement: Optional[Any]
    textGiftElement: Optional[Any]
    videoElement: Optional[Any]
    walletElement: Optional[Any]
    yoloGameResultElement: Optional[Any]


class ChatType(IntEnum):
    FRIEND = 1
    GROUP = 2


class OtherAdd(BaseModel):
    uid: Optional[str]
    name: Optional[str]
    uin: Optional[str]


class MemberAdd(BaseModel):
    showType: int
    otherAdd: Optional[OtherAdd]
    otherAddByOtherQRCode: Optional[Any]
    otherAddByYourQRCode: Optional[Any]
    youAddByOtherQRCode: Optional[Any]
    otherInviteOther: Optional[Any]
    otherInviteYou: Optional[Any]
    youInviteOther: Optional[Any]


class ShutUpTarget(BaseModel):
    uid: Optional[str]
    card: str
    name: str
    role: int
    uin: str


class ShutUp(BaseModel):
    curTime: int
    duration: int
    admin: ShutUpTarget
    member: ShutUpTarget


class GroupElement(BaseModel):
    type: int
    role: int
    groupName: Optional[str]
    memberUid: Optional[str]
    memberNick: Optional[str]
    memberRemark: Optional[str]
    adminUid: Optional[str]
    adminNick: Optional[str]
    adminRemark: Optional[str]
    createGroup: Optional[Any]
    memberAdd: Optional[MemberAdd]
    shutUp: Optional[ShutUp]
    memberUin: Optional[str]
    adminUin: Optional[str]


class XmlElement(BaseModel):
    busiType: Optional[str]
    busiId: Optional[str]
    c2cType: int
    serviceType: int
    ctrlFlag: int
    content: Optional[str]
    templId: Optional[str]
    seqId: Optional[str]
    templParam: Optional[Any]
    pbReserv: Optional[str]
    members: Optional[Any]


class Member(BaseModel):
    uid: str
    qid: str
    uin: str
    nick: str
    remark: str
    cardType: int
    cardName: str
    role: int
    avatarPath: str
    shutUpTime: int
    isDelete: bool


class Group(BaseModel):
    groupCode: str
    maxMember: int
    memberCount: int
    groupName: str
    groupStatus: int
    memberRole: int
    isTop: bool
    toppedTimestamp: str
    privilegeFlag: int
    isConf: bool
    hasModifyConfGroupFace: bool
    hasModifyConfGroupName: bool
    remarkName: str
    avatarUrl: str
    hasMemo: bool
    groupShutupExpireTime: str
    personShutupExpireTime: str
    discussToGroupUin: str
    discussToGroupMaxMsgSeq: int
    discussToGroupTime: int


class ImageInfo(BaseModel):
    width: int
    height: int
    type: Optional[str]
    mime: Optional[str]
    wUnits: Optional[str]
    hUnits: Optional[str]


class UploadResponse(BaseModel):
    md5: str
    imageInfo: Optional[ImageInfo]
    fileSize: int
    filePath: str
    ntFilePath: str


class MsgType(IntEnum):
    normal = 2
    may_file = 3
    system = 5
    voice = 6
    video = 7
    value8 = 8
    reply = 9
    wallet = 10
    ark = 11
    may_market = 17


class MessageEvent(RedEvent):
    """消息事件"""

    __event__ = "message"
    post_type: Literal["message"]
    message_type: Literal["private", "group"]
    sub_type: Union[Literal["channel"], str]
    message: RedMessage
    original_message: RedMessage

    def __repr__(self) -> str:
        return f'Event<{self.type}>: "{self.message}"'

    async def reply(self, msg: "T_RedMSG") -> Dict[str, Any]:
        """回复消息"""

        raise NotImplementedError


class PrivateMessageEvent(MessageEvent):
    """私聊消息事件"""

    __event__ = "message.private"
    message_type: Literal["private"]
    sub_type: Literal["friend", "group", "group_self", "other"]
    roleType: int

    async def reply(self, msg: T_RedMSG) -> Dict[str, Any]:
        return await self.adapter.send_message(
            chatType=1, peerUin=self.peerUid, elements=RedMessage(msg)
        )


class GroupMessageEvent(MessageEvent):
    """群消息事件"""

    __event__ = "message.group"
    message_type: Literal["group"]
    sub_type: Literal["normal", "anonymous", "notice"]
    roleType: int

    async def reply(self, msg: T_RedMSG) -> Dict[str, Any]:
        return await self.adapter.send_message(
            chatType=2, peerUin=self.peerUid, elements=RedMessage(msg)
        )


class NoticeEvent(RedEvent):
    __event__ = "notice"
    post_type: Literal["notice"]
    notice_type: str
    msgId: str
    msgRandom: str
    msgSeq: str
    cntSeq: str
    chatType: ChatType
    msgType: MsgType
    subMsgType: int
    peerUid: str
    peerUin: Optional[str]

    # class Config:
    #     extra = "ignore"


class GroupNameUpdateEvent(NoticeEvent):
    """群名变更事件"""

    __event__ = "notice.group_name_update"
    notice_type: Literal["group_name_update"]
    currentName: str
    operatorUid: str
    operatorName: str


class MemberAddEvent(NoticeEvent):
    """群成员增加事件"""

    __event__ = "notice.member_add"
    notice_type: Literal["member_add"]
    memberUid: str
    operatorUid: str
    memberName: Optional[str]


class MemberMuteEvent(NoticeEvent):
    """群成员禁言相关事件"""

    __event__ = "notice.member_mute"
    notice_type: Literal["member_mute"]
    start: datetime
    duration: timedelta
    operator: ShutUpTarget
    member: ShutUpTarget


class MemberUnmuteEvent(NoticeEvent):
    """群成员被解除禁言事件"""

    __event__ = "notice.member_unmute"
    notice_type: Literal["member_unmute"]
    start: datetime
    duration: timedelta
    operator: ShutUpTarget
    member: ShutUpTarget


class MetaEvent(RedEvent):
    """元事件"""

    __event__ = "meta_event"
    post_type: Literal["meta_event"]
    meta_event_type: str


class LifecycleMetaEvent(MetaEvent):
    """生命周期"""

    __event__ = "meta_event.lifecycle"
    meta_event_type: Literal["lifecycle"]
    sub_type: Literal["enable", "disable", "connect"]


_red_events = {
    model.__event__: model
    for model in globals().values()
    if inspect.isclass(model) and issubclass(model, RedEvent)
}


def get_event_class(
    post_type: str, event_type: str, sub_type: Optional[str] = None
) -> Type[T_RedEvent]:
    """根据接收到的消息类型返回对应的事件类。

    Args:
        post_type: 请求类型。
        event_type: 事件类型。
        sub_type: 子类型。

    Returns:
        对应的事件类。
    """
    if sub_type is None:
        return _red_events[".".join((post_type, event_type))]
    return (
        _red_events.get(".".join((post_type, event_type, sub_type)))
        or _red_events[".".join((post_type, event_type))]
    )



"""APScheduler 适配器配置。"""

from typing import Any, Dict

from pydantic import Field

from iamai.config import ConfigModel

__all__ = ["Config"]


class Config(ConfigModel):
    """APScheduler 配置类，将在适配器被加载时被混入到机器人主配置中。

    Attributes:
        scheduler_config: 调度器配置。
    """

    __config_name__ = "apscheduler"
    scheduler_config: Dict[str, Any] = Field(default_factory=dict)



from iamai import Plugin


class Bradge(Plugin):
    async def handle(self) -> None:
        if self.event.adapter.name == "kook":
            await self.bot.get_adapter("cqhttp").call_api(
                "send_group_msg",
                group_id=971050440,  # 替换为你想要的目标转发群号
                message=f"[{self.event.adapter.name} - {self.event.extra.author.username}]\n{self.event.message}",
            )
        elif self.event.adapter.name == "cqhttp":
            if self.event.group_id == 971050440:  # 替换为你想要的目标消息接收群号
                await self.bot.get_adapter("kook").call_api(
                    api="message/create",
                    target_id=1661426334688259,  # 替换为你想要的频道Channel号
                    content=f"[{self.event.adapter.name} - {self.event.sender.nickname}]\n{self.event.message}",
                )

    async def rule(self) -> bool:
        if self.event.adapter.name not in ["cqhttp", "kook"]:
            return False
        return self.event.type in ["message", 9]



"""iamai 协议适配器。

所有协议适配器都必须继承自 `Adapter` 基类。
"""

import os
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Generic,
    Optional,
    Type,
    TypeVar,
    Union,
    final,
    overload,
)

from ..event import Event
from ..typing import ConfigT, EventT
from ..utils import is_config_class

if TYPE_CHECKING:
    from ..bot import Bot

__all__ = ["Adapter"]

if os.getenv("IAMAI_DEV") == "1":  # pragma: no cover
    # 当处于开发环境时，使用 pkg_resources 风格的命名空间包
    __import__("pkg_resources").declare_namespace(__name__)


_EventT = TypeVar("_EventT", bound="Event[Any]")


class Adapter(Generic[EventT, ConfigT], ABC):
    """协议适配器基类。

    Attributes:
        name: 适配器的名称。
        bot: 当前的机器人对象。
    """

    name: str
    bot: "Bot"
    Config: Type[ConfigT]

    def __init__(self, bot: "Bot") -> None:
        """初始化。

        Args:
            bot: 当前机器人对象。
        """
        if not hasattr(self, "name"):
            self.name = self.__class__.__name__
        self.bot: Bot = bot
        self.handle_event = self.bot.handle_event

    @property
    def config(self) -> ConfigT:
        """适配器配置。"""
        default: Any = None
        config_class = getattr(self, "Config", None)
        if is_config_class(config_class):
            return getattr(
                self.bot.config.adapter,
                config_class.__config_name__,
                default,
            )
        return default

    @final
    async def safe_run(self) -> None:
        """附带有异常处理地安全运行适配器。"""
        try:
            await self.run()
        except Exception as e:
            self.bot.error_or_exception(
                f"Run adapter {self.__class__.__name__} failed:", e
            )

    @abstractmethod
    async def run(self) -> None:
        """适配器运行方法，适配器开发者必须实现该方法。

        适配器运行过程中保持保持运行，当此方法结束后， AliceBot 不会自动重新启动适配器。
        """
        raise NotImplementedError

    async def startup(self) -> None:
        """在适配器开始运行前运行的方法，用于初始化适配器。

        AliceBot 依次运行并等待所有适配器的 `startup()` 方法，待运行完毕后再创建 `run()` 任务。
        """

    async def shutdown(self) -> None:
        """在适配器结束运行时运行的方法，用于安全地关闭适配器。

        AliceBot 在接收到系统的结束信号后依次运行并等待所有适配器的 `shutdown()` 方法。
        当强制退出时此方法可能未被执行。
        """

    @overload
    async def get(
        self,
        func: Optional[Callable[[EventT], Union[bool, Awaitable[bool]]]] = None,
        *,
        event_type: None = None,
        max_try_times: Optional[int] = None,
        timeout: Optional[Union[int, float]] = None,
    ) -> EventT: ...

    @overload
    async def get(
        self,
        func: Optional[Callable[[_EventT], Union[bool, Awaitable[bool]]]] = None,
        *,
        event_type: Type[_EventT],
        max_try_times: Optional[int] = None,
        timeout: Optional[Union[int, float]] = None,
    ) -> _EventT: ...

    @final
    async def get(
        self,
        func: Optional[Callable[[Any], Union[bool, Awaitable[bool]]]] = None,
        *,
        event_type: Any = None,
        max_try_times: Optional[int] = None,
        timeout: Optional[Union[int, float]] = None,
    ) -> Event[Any]:
        """获取满足指定条件的的事件，协程会等待直到适配器接收到满足条件的事件、超过最大事件数或超时。

        类似 `Bot` 类的 `get()` 方法，但是隐含了判断产生事件的适配器是本适配器。
        等效于 `Bot` 类的 `get()` 方法传入 adapter_type 为本适配器类型。

        Args:
            func: 协程或者函数，函数会被自动包装为协程执行。
                要求接受一个事件作为参数，返回布尔值。
                当协程返回 `True` 时返回当前事件。
                当为 `None` 时相当于输入对于任何事件均返回真的协程，即返回适配器接收到的下一个事件。
            event_type: 当指定时，只接受指定类型的事件，先于 func 条件生效。默认为 `None`。
            max_try_times: 最大事件数。
            timeout: 超时时间。

        Returns:
            返回满足 func 条件的事件。

        Raises:
            GetEventTimeout: 超过最大事件数或超时。
        """
        return await self.bot.get(
            func,
            event_type=event_type,
            adapter_type=type(self),
            max_try_times=max_try_times,
            timeout=timeout,
        )



from iamai import Plugin


class GlobalStateTest2(Plugin):
    async def handle(self) -> None:
        if self.bot.global_state.get("count", None) is None:
            self.bot.global_state["count"] = 0
        self.bot.global_state["count"] -= 1
        await self.event.reply(f'sub: {self.bot.global_state["count"]}')

    async def rule(self) -> bool:
        if self.event.adapter.name != "cqhttp":
            return False
        if self.event.type != "message":
            return False
        return self.event.message.get_plain_text() == "sub"



from aiohttp import web

from iamai import Plugin
from iamai.log import logger

from .config import EVENT_DESCRIPTIONS

server = None
app = web.Application()


class F(Plugin):
    async def handle(self) -> None:
        global server

        if str(self.event.message) == "service on":
            app.router.add_post("/", self.handle_request)
            runner = web.AppRunner(app)
            await runner.setup()
            site = web.TCPSite(runner, "localhost", 8888)

            try:
                await site.start()
                logger.info("Server started on localhost:8888")
                await self.event.reply("Server started on localhost:8888")
            except OSError:
                logger.error("Failed to start server. Port is already in use.")
                await self.event.reply(
                    "Failed to start server. Port is already in use."
                )
        elif str(self.event.message) == "service off":
            if server is not None:
                server.close()
                await server.wait_closed()
                await self.event.reply("Server Off.")
            else:
                await self.event.reply("Server is not running.")

    async def handle_request(self, request):
        data = await request.json()
        # logger.info(f'Received JSON body: {data}')

        event_type = request.headers.get("X-GitHub-Event")
        if event_type in [
            "commit_comment",
            "create",
            "delete",
            "fork",
            "issue_comment",
            "issues",
            "pull_request",
            "push",
            "release",
            "watch",
        ]:
            message = self._format_event(event_type=event_type, data=data)
            if message:
                await self.event.adapter.call_api(
                    "send_group_msg", group_id=126211793, message=message
                )
            else:
                return

        response = {"message": "Received request"}
        return web.json_response(response)

    async def rule(self) -> bool:
        return str(self.event.message) in ["service on", "service off"]

    @staticmethod
    def _format_event(self, event_type, data):
        try:
            if isinstance(EVENT_DESCRIPTIONS[event_type], dict):
                return EVENT_DESCRIPTIONS[event_type][data["action"]].format(**data)
            elif isinstance(EVENT_DESCRIPTIONS[event_type], str):
                return EVENT_DESCRIPTIONS[event_type].format(**data)
        except KeyError:
            return None


"""
>smee -u https://smee.io/MHldOrjI8msgNUa5 -p 8888
"""



from .model import *  # noqa: F403

api_method_map = {
    "asset/create": {"method": "POST", "type": URL},  # noqa: F405
    "blacklist/create": {"method": "POST", "type": None},
    "blacklist/delete": {"method": "POST", "type": None},
    "blacklist/list": {"method": "GET", "type": BlackListsReturn},  # noqa: F405
    "channel-role/create": {"method": "POST", "type": ChannelRoleReturn},  # noqa: F405
    "channel-role/delete": {"method": "POST", "type": None},
    "channel-role/index": {"method": "GET", "type": ChannelRoleInfo},  # noqa: F405
    "channel-role/update": {"method": "POST", "type": ChannelRoleReturn},  # noqa: F405
    "channel/create": {"method": "POST", "type": Channel},  # noqa: F405
    "channel/delete": {"method": "POST", "type": None},
    "channel/update": {"method": "POST", "type": Channel},  # noqa: F405
    "channel/list": {"method": "GET", "type": ChannelsReturn},  # noqa: F405
    "channel/move-user": {"method": "POST", "type": None},
    "channel/user-list": {"method": "POST", "type": List[User]},  # noqa: F405
    "channel/view": {"method": "GET", "type": Channel},  # noqa: F405
    "direct-message/add-reaction": {"method": "POST", "type": None},
    "direct-message/create": {"method": "POST", "type": MessageCreateReturn},  # noqa: F405
    "direct-message/delete": {"method": "POST", "type": None},
    "direct-message/delete-reaction": {"method": "POST", "type": None},
    "direct-message/list": {"method": "GET", "type": DirectMessagesReturn},  # noqa: F405
    "direct-message/reaction-list": {"method": "GET", "type": List[ReactionUser]},  # noqa: F405
    "direct-message/update": {"method": "POST", "type": None},
    "direct-message/view": {"method": "GET", "type": DirectMessage},  # noqa: F405
    "gateway/index": {"method": "GET", "type": URL},  # noqa: F405
    "guild-emoji/create": {"method": "POST", "type": None},
    "guild-emoji/delete": {"method": "POST", "type": None},
    "guild-emoji/list": {"method": "GET", "type": GuildEmojisReturn},  # noqa: F405
    "guild-emoji/update": {"method": "POST", "type": None},
    "guild-mute/create": {"method": "POST", "type": None},
    "guild-mute/delete": {"method": "POST", "type": None},
    "guild-mute/list": {"method": "GET", "type": None},
    "guild-role/create": {"method": "POST", "type": Role},  # noqa: F405
    "guild-role/delete": {"method": "POST", "type": None},
    "guild-role/grant": {"method": "POST", "type": GuilRoleReturn},  # noqa: F405
    "guild-role/list": {"method": "GET", "type": RolesReturn},  # noqa: F405
    "guild-role/revoke": {"method": "POST", "type": GuilRoleReturn},  # noqa: F405
    "guild-role/update": {"method": "POST", "type": Role},  # noqa: F405
    "guild/kickout": {"method": "POST", "type": None},
    "guild/leave": {"method": "POST", "type": None},
    "guild/list": {"method": "GET", "type": GuildsReturn},  # noqa: F405
    "guild/nickname": {"method": "POST", "type": None},
    "guild/user-list": {"method": "GET", "type": GuildUsersRetrun},  # noqa: F405
    "guild/view": {"method": "GET", "type": Guild},  # noqa: F405
    "intimacy/index": {"method": "GET", "type": IntimacyIndexReturn},  # noqa: F405
    "intimacy/update": {"method": "POST", "type": None},
    "invite/create": {"method": "POST", "type": URL},  # noqa: F405
    "invite/delete": {"method": "POST", "type": None},
    "invite/list": {"method": "GET", "type": InvitesReturn},  # noqa: F405
    "message/add-reaction": {"method": "POST", "type": None},
    "message/create": {"method": "POST", "type": MessageCreateReturn},  # noqa: F405
    "message/delete": {"method": "POST", "type": None},
    "message/delete-reaction": {"method": "POST", "type": None},
    "message/list": {"method": "GET", "type": ChannelMessagesReturn},  # noqa: F405
    "message/reaction-list": {"method": "GET", "type": List[ReactionUser]},  # noqa: F405
    "message/update": {"method": "POST", "type": None},
    "message/view": {"method": "GET", "type": ChannelMessage},  # noqa: F405
    "user-chat/create": {"method": "POST", "type": UserChat},  # noqa: F405
    "user-chat/delete": {"method": "POST", "type": None},
    "user-chat/list": {"method": "GET", "type": UserChatsReturn},  # noqa: F405
    "user-chat/view": {"method": "GET", "type": UserChat},  # noqa: F405
    "user/me": {"method": "GET", "type": User},  # noqa: F405
    "user/offline": {"method": "POST", "type": None},
    "user/view": {"method": "GET", "type": User},  # noqa: F405
}


def get_api_method(api: str) -> str:
    return api_method_map.get(api, {}).get("method", "POST")


def get_api_restype(api: str) -> Any:  # noqa: F405
    return api_method_map.get(api, {}).get("type")



"""APScheduler 适配器。

本适配器用于实现定时任务，适配器将使用 APScheduler 实现定时任务，在设定的时间产生一个事件供插件处理。
APScheduler 使用方法请参考：[APScheduler](https://apscheduler.readthedocs.io/)。
"""

import inspect
from functools import wraps
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Type, Union

from apscheduler.job import Job
from apscheduler.schedulers.asyncio import AsyncIOScheduler

from iamai.adapter import Adapter
from iamai.log import logger
from iamai.plugin import Plugin
from iamai.typing import PluginT

from .config import Config
from .event import APSchedulerEvent

if TYPE_CHECKING:
    from apscheduler.triggers.base import BaseTrigger

__all__ = ["APSchedulerAdapter", "scheduler_decorator"]


class APSchedulerAdapter(Adapter[APSchedulerEvent, Config]):
    """APScheduler 适配器。"""

    name: str = "apscheduler"
    Config = Config

    scheduler: AsyncIOScheduler
    plugin_class_to_job: Dict[Type[Plugin[Any, Any, Any]], Job]

    async def startup(self) -> None:
        """创建 `AsyncIOScheduler` 对象。"""
        self.scheduler = AsyncIOScheduler(self.config.scheduler_config)
        self.plugin_class_to_job = {}

    async def run(self) -> None:
        """启动调度器。"""
        for plugin in self.bot.plugins:
            if not hasattr(plugin, "__schedule__"):
                continue

            if not hasattr(plugin, "trigger") or not hasattr(plugin, "trigger_args"):
                logger.error(
                    f"Plugin {plugin.__name__} __schedule__ is True, "
                    f"but did not set trigger or trigger_args"
                )
                continue

            trigger: Union[str, BaseTrigger] = getattr(plugin, "trigger")  # noqa: B009
            trigger_args: Dict[str, Any] = getattr(plugin, "trigger_args")  # noqa: B009

            if not isinstance(trigger, str) or not isinstance(trigger_args, dict):
                logger.error(
                    f"Plugin {plugin.__name__} trigger or trigger_args type error"
                )
                continue

            try:
                self.plugin_class_to_job[plugin] = self.scheduler.add_job(
                    self.create_event, args=(plugin,), trigger=trigger, **trigger_args
                )
            except Exception as e:
                self.bot.error_or_exception(
                    f"Plugin {plugin.__name__} add_job filed, "
                    "please check trigger and trigger_args:",
                    e,
                )
            else:
                logger.info(f"Plugin {plugin.__name__} has been scheduled to run")

        self.scheduler.start()

    async def shutdown(self) -> None:
        """关闭调度器。"""
        self.scheduler.shutdown()

    async def create_event(self, plugin_class: Type[Plugin[Any, Any, Any]]) -> None:
        """创建 `APSchedulerEvent` 事件。

        Args:
            plugin_class: `Plugin` 类。
        """
        logger.info(f"APSchedulerEvent set by {plugin_class} is created as scheduled")
        await self.handle_event(
            APSchedulerEvent(adapter=self, plugin_class=plugin_class),
            handle_get=False,
            show_log=False,
        )

    async def send(self, *args: Any, **kwargs: Any) -> Any:
        """APScheduler 适配器不适用发送消息。"""
        raise NotImplementedError


def scheduler_decorator(
    trigger: str, trigger_args: Dict[str, Any], override_rule: bool = False
) -> Callable[[Type[PluginT]], Type[PluginT]]:
    """用于为插件类添加计划任务功能的装饰器。

    Args:
        trigger: APScheduler 触发器。
        trigger_args: APScheduler 触发器参数。
        override_rule: 是否重写 `rule()` 方法。
            若为 `True`，则会在 `rule()` 方法中添加处理本插件定义的计划任务事件的逻辑。
    """

    def _decorator(cls: Type[PluginT]) -> Type[PluginT]:
        if not inspect.isclass(cls):
            raise TypeError("can only decorate class")
        if not issubclass(cls, Plugin):
            raise TypeError("can only decorate Plugin class")
        setattr(cls, "__schedule__", True)  # noqa: B010
        setattr(cls, "trigger", trigger)  # noqa: B010
        setattr(cls, "trigger_args", trigger_args)  # noqa: B010
        if override_rule:

            def _rule_decorator(func: Callable[[PluginT], Awaitable[bool]]) -> Any:
                @wraps(func)
                async def _wrapper(self: PluginT) -> bool:
                    if (
                        self.event.type == "apscheduler"
                        # pylint: disable-next=unidiomatic-typecheck
                        and type(self) is self.event.plugin_class
                    ):
                        return True
                    return await func(self)

                return _wrapper

            cls.rule = _rule_decorator(cls.rule)  # type: ignore
        return cls  # type: ignore

    return _decorator



from iamai import Plugin


class Echo(Plugin):
    async def handle(self) -> None:
        await self.event.reply(self.event.message.replace("echo ", ""))

    async def rule(self) -> bool:
        if self.event.adapter.name != "cqhttp":
            return False
        if self.event.type != "message_sent":
            return False
        return self.event.message.startswith("echo ")



"""CQHTTP 适配器异常。"""

from typing import Any, ClassVar, Dict

from iamai.exceptions import AdapterException

__all__ = [
    "CQHTTPException",
    "NetworkError",
    "ActionFailed",
    "ApiNotAvailable",
    "ApiTimeout",
]


class CQHTTPException(AdapterException):
    """CQHTTP 异常基类。"""


class NetworkError(CQHTTPException):
    """网络异常。"""


class ActionFailed(CQHTTPException):
    """API 请求成功响应，但响应表示 API 操作失败。"""

    def __init__(self, resp: Dict[str, Any]) -> None:
        """初始化。

        Args:
            resp: 返回的响应。
        """
        self.resp = resp


class ApiNotAvailable(ActionFailed):
    """API 请求返回 404，表示当前请求的 API 不可用或不存在。"""

    ERROR_CODE: ClassVar[int] = 1404


class ApiTimeout(CQHTTPException):
    """API 请求响应超时。"""



from iamai import Plugin


class Hallo(Plugin):
    async def handle(self) -> None:
        await self.event.reply("Hello!")

    async def rule(self) -> bool:
        if self.event.adapter.name != "gensokyo":
            return False
        if self.event.type != "message":
            return False
        return str(self.event.message).lower() == "shiki"



from iamai import Plugin


class Reload(Plugin):
    async def handle(self) -> None:
        self.bot.reload_plugins()
        await self.event.reply("\n".join(map(repr, self.bot.plugins)))

    async def rule(self) -> bool:
        if self.event.adapter.name != "cqhttp":
            return False
        if self.event.type != "message_sent":
            return False
        return self.event.message.get_plain_text() == "reload"



"""iamai event.

Base class for event classes. Adapter developers should implement a subclass of this event class base class.
"""

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, Optional, Union
from typing_extensions import Self

from pydantic import BaseModel, ConfigDict

from .typing import AdapterT

__all__ = ["Event", "MessageEvent"]


class Event(ABC, BaseModel, Generic[AdapterT]):
    """The base class of event classes.

    Attributes:
        adapter: The adapter object that generated the current event.
        type: event type.
        __handled__: Indicates whether the event has been handled, used for adapter processing. Warning: Do not change the value of this property manually.
    """

    model_config = ConfigDict(extra="allow")

    if TYPE_CHECKING:
        adapter: AdapterT
    else:
        adapter: Any
    type: Optional[str]
    __handled__: bool = False

    def __str__(self) -> str:
        """Returns the textual representation of the event.

        Returns:
            A textual representation of the event.
        """
        return f"Event<{self.type}>"

    def __repr__(self) -> str:
        """Returns the description of the event.

        Returns:
            Description of the event.
        """
        return self.__str__()


class MessageEvent(Event[AdapterT], Generic[AdapterT]):
    """Base class for general message event classes."""

    @abstractmethod
    def get_plain_text(self) -> str:
        """Get the plain text content of the message.

        Returns:
            The plain text content of the message.
        """

    @abstractmethod
    async def reply(self, message: str) -> Any:
        """Reply message.

        Args:
            message: The content of the reply message.

        Returns:
            The response to the reply message action.
        """

    @abstractmethod
    async def is_same_sender(self, other: Self) -> bool:
        """Determine whether itself and another event are the same sender.

        Args:
            other: another event.

        Returns:
            Is it the same sender?
        """

    async def get(
        self,
        *,
        max_try_times: Optional[int] = None,
        timeout: Optional[Union[int, float]] = None,
    ) -> Self:
        """Get the user's reply message.

        Equivalent to `get()` of ``Bot``, the condition is that the adapter, event type and sender are the same.

        Args:
            max_try_times: Maximum number of events.
            timeout: timeout period.

        Returns:
            Message event that the user replies to.

        Raises:
            GetEventTimeout: Maximum number of events exceeded or timeout.
        """

        return await self.adapter.get(
            self.is_same_sender,
            event_type=type(self),
            max_try_times=max_try_times,
            timeout=timeout,
        )

    async def ask(
        self,
        message: str,
        max_try_times: Optional[int] = None,
        timeout: Optional[Union[int, float]] = None,
    ) -> Self:
        """Ask for news.

        Indicates getting the user's reply after replying to a message.
        Equivalent to executing ``get()`` after ``reply()``.

        Args:
            message: The content of the reply message.
            max_try_times: Maximum number of events.
            timeout: timeout period.

        Returns:
            Message event that the user replies to.
        """

        await self.reply(message)
        return await self.get(max_try_times=max_try_times, timeout=timeout)
